#!/usr/bin/env python3

"""
Geoscience Australia - Python Geodesy Package
Survey Module
"""

from math import sqrt, sin, cos, atan, radians, degrees, exp
from statistics import mean, stdev
from geodepy.convert import rect2polar, polar2rect


def first_vel_params(wavelength, frequency, n_REF=None, unit_length=None):
    """
    Calculates the constant First Velocity Correction Parameters C and D
    for a given set of standard instrument settings
    :param wavelength: Instrument Carrier Wavelength (Micrometers) - Mandatory
    :param frequency: Instrument modulation frequency (Hz) - Optional
    :param n_REF: manufacturers reference refractive index - Recommended
    :param unit_length: unit length of instrument - Optional
    :return: First Velocity Correction Parameters C and D
    Reference Rueger, J.M., 2012, Electronic Distance Measurement – An Introduction, 4rd edition, Springer, Berlin
    """
    
    if not n_REF :
        if not unit_length:
            raise ValueError('Error - n_REF and unit_length cannot both be None')
        if not frequency:
            raise ValueError('Error - n_REF and frequency cannot both be None')
        # Rueger eq 6.3
        n_REF = 299792458 / (2 * unit_length * frequency)
    
    # Rueger eq 6.12
    param_c = (n_REF-1)*10**6
    
    # Rueger eq 5.12
    # https://office.iag-aig.org/doc/5d7b8fda0c032.pdf Resolution 3
    nG_1_10_6 = 287.6155 + (4.8866 / wavelength ** 2) + (0.068 / wavelength ** 4)
    # Rueger eq 6.13
    param_d = (273.15 / 1013.25) * nG_1_10_6
    
    return param_c, param_d


def part_h2o_vap_press(dry_temp, pressure, rel_humidity=None, wet_temp=None):
    """
    :param dry_temp: Observed Dry Temperature (degrees Celsius)
    :param pressure: Observed Pressure (hectopascals or millibars)
    :param rel_humidity: Observed Relative Humidity (percentage) - Optional if wet_temp supplied
    :param wet_temp: Observed Wet Temperature (degrees Celsius) - Optional if rel_humidity supplied
    Reference Rueger, J.M., 2012, Electronic Distance Measurement – An Introduction, 4rd edition, Springer, Berlin
    """
    
    if not rel_humidity and not wet_temp:
        raise ValueError('Error - rel_humidity and wet_temp cannot both be None')
                         
    if rel_humidity:
        wet_temp = dry_temp
    # Rueger eq 5.27
    E_w = ((1.0007 + ((3.46 * pressure) * (10**-6)))
         * 6.1121 * exp((17.502 * wet_temp) / (240.94 + wet_temp)))
    
    if rel_humidity:
        # Rueger eq 5.29
        e = (E_w*rel_humidity)/100
    else:
        e = E_w - 0.000662 * pressure * (dry_temp - wet_temp)
    
    return e
    

def first_vel_corrn(dist, first_vel_param, temp, pressure,
                    rel_humidity=None, wet_temp=None, CO2_ppm=None, 
                    wavelength=None):
    """
    Carries out First Velocity Correction of Electronic Distance Measurement,
    given correction parameters and atmospheric observations
    :param dist: Uncorrected Observed Slope Distance
    :param first_vel_param: Tuple of First Velocity Parameters C and D (see function first_vel_params)
    :param temp: Observed Temperature (degrees Celsius)
    :param pressure: Observed Pressure (hectopascals or millibars)
    :param rel_humidity: Observed Relative Humidity (percentage)
    :return: Slope Distance with First Velocity Correction applied
    """
    param_c = first_vel_param[0]
    if not CO2_ppm:    
        param_d = first_vel_param[1]
        e = part_h2o_vap_press(temp, pressure, rel_humidity, wet_temp)
        
        # Rueger eq 6.11
        t_273_15 = temp+273.15
        first_vel_corrn_ppm = (param_c-
                               ((param_d*pressure)/t_273_15)+
                               ((11.27*e)/t_273_15))
        first_vel_corrn_metres = dist * first_vel_corrn_ppm * (10 ** -6)
        
    else:
        if all([temp, pressure, rel_humidity, wavelength]):
            e = humidity2part_water_vapour_press(rel_humidity, temp)
            NPROPG_1 = group_refractivity(
                wavelength, temp, pressure, e, CO2_ppm)
        else:
            raise ValueError('Error - temp, pressure, rel_humidity and wavelength '+
                             'are all mandatory when CO2_ppm is not None')
        
        n_ref = 1 + (param_c / 1.0e6)
        n_g = 1 + (NPROPG_1 / 1.0e8)
        first_vel_corrn_metres = ((n_ref / n_g)-1) * dist
        
    return first_vel_corrn_metres


def mets_partial_differentials(group_ref_Index=1.00028, 
                               temp=15, 
                               pressure=1013.25,
                               rel_humidity=60):
    """ Returns the sensitivity coefficients for temp, pressure and humidity in ppm
    :param group_ref_Index: manufacturers group refractive index of light
    :param temp: Observed Dry Temperature (degrees Celsius)
    :param pressure: Observed Pressure (hectopascals or millibars)
    :param rel_humidity: Observed Relative Humidity (percentage) - Optional if wet_temp supplied
    """
    ng_1 = group_ref_Index - 1
    t_273_15 = temp + 273.15
    e = part_h2o_vap_press(temp, pressure, rel_humidity)
    
    param_k = (((((ng_1 * 273.15 * pressure)
                  /1013.25)
                 -(11.27 * e * 10**-6))
                /t_273_15**2)
               *(10**6))
    param_l = (0.26957809 * (ng_1 / t_273_15)
               *(10**6))
    param_m = (((11.27 * 10**-6)
               /t_273_15)
               *(10**6))
    
    return param_k, param_l, param_m


def precise_inst_ht(vert_list, spacing, offset):
    """
    Uses a set of Vertical Angle Observations taken to a
    levelling staff at regular intervals to determine the
    height of the instrument above a reference mark
    :param vert_list: List of Vertical (Zenith) Angle Observations (minimum of 3) in Decimal Degrees format
    :param spacing: Distance in metres between each vertical angle observation
    :param offset: Lowest observed height above reference mark
    :return: Instrument Height above reference mark and its standard deviation
    """
    if len(vert_list) < 3:
        raise ValueError('ValueError: 3 or more vertical angles required')
    vert_list.sort(reverse=True)
    vert_pairs = [(va1, va2) for va1, va2 in zip(vert_list, vert_list[1:])]
    base_ht = []
    height_comp = []
    for num, pair in enumerate(vert_pairs):
        base_ht_pair = offset + num * spacing
        base_ht.append(base_ht_pair)
        dist_a = sin(radians(pair[1])) * (spacing / (sin(radians(pair[0] - pair[1]))))
        delta_ht = dist_a * (sin(radians(pair[0] - 90)))
        height_comp.append(delta_ht + base_ht[num])
    return round(mean(height_comp), 5), round(stdev(height_comp), 5)


def joins(east1, north1, east2, north2):
    return rect2polar(east2 - east1, north2 - north1)


def radiations(east1, north1, brg1to2, dist, rotation=0, psf=1):
    delta_east, delta_north = polar2rect(dist * psf, brg1to2 + rotation)
    return east1 + delta_east, north1 + delta_north


def va_conv(zenith_angle, slope_dist, height_inst=0, height_tgt=0):
    """
    Function to convert vertical angles (zenith distances) and slope distances
    into horizontal distances and changes in height. Instrument and Target
    heights can be entered to allow computation of zenith and slope distances
    between ground points.

    :param zenith_angle:    Zenith Angle from Instrument to Target, expressed
                            in decimal degrees
    :param slope_dist:      Slope Distance from Instrument to Target in metres
    :param height_inst:     Height of Instrument. Optional - Default Value of 0m
    :param height_tgt:      Height of Target. Optional - Default Value of 0m

    :return: vert_angle_pt: Vertical Angle between Ground Points, expressed
                            in decimal degrees
    :return: slope_dist_pt: Slope Distance between Ground Points in metres
    :return: hz_dist:       Horizontal Distance
    :return: delta_ht:      Change in height between Ground Points in metres
    """
    # Convert Zenith Angle to Vertical Angle
    try:
        if zenith_angle == 0 or zenith_angle == 180:
            raise ValueError
        elif 0 < zenith_angle < 180:
            zenith_angle = radians(90 - zenith_angle)
        elif 180 < zenith_angle < 360:
            zenith_angle = radians(270 - zenith_angle)
        else:
            raise ValueError
    except ValueError:
        raise ValueError('ValueError: Vertical Angle Invalid')
    # Calculate Horizontal Dist and Delta Height
    hz_dist = slope_dist * cos(zenith_angle)
    delta_ht = slope_dist * sin(zenith_angle)
    # Account for Target and Instrument Heights
    delta_ht = height_inst + delta_ht - height_tgt
    slope_dist_pt = sqrt(delta_ht ** 2 + hz_dist ** 2)
    vert_angle_pt = atan(delta_ht / hz_dist)
    vert_angle_pt = degrees(vert_angle_pt)
    return vert_angle_pt, slope_dist_pt, hz_dist, delta_ht

"""
The following has been translated from UNSW FORTRAN code
"""
"""
    PROGRAM RINDEX

    -----------------------------------------------------------------
    Date :  6 July 1995

    Version 4.0 with CF = 1.022

    This program calculates the phase and group refractive index
    using equations for the visible and near infrared developed by
    P.E. Ciddor of the National Measurement Laboratory, CSIRO,
    Division of the Applied Physics, Australia. (see reference below)

    Program originally written by S.K. Johnson and J.M. R?eger of the
    School of Geomatic Engineering, University of New South
    Wales, Australia.
    Program converted from FORTRAN to python by Kent Wheeler of Landgate.

    Main References:
    P.E. Ciddor, 1996, Refractive Index of Air: New Equations for the 
          Visible and Near Infrared, Applied Optics (Lasers, Phototonics,
          and Environmental optics), 35(9): 1566-73.
    R.S. Davies, 1992, Equations for Determination of the Density of Moist 
          Air (1981/91), Metrologia, 29: 67-70.
    P. Giacomo, 1982, Equation for Determination of the Density of Moist 
          Air, Metrologia, 18: 33-40.
    E.P. Peck and K. Reeder, 1972, Dispersion of Air, Journal of the
          Optical Society of America, 62(8): 958-962.
    J.C. Owens, 1967, Optical Refractive Index of Air: Dependence on
          Pressure, Temperature and Composition, Applied Optics, 6(1): 51-59
    K.P Birch and M.J. Downs, 1994, Correction to the updated Edlen Equation
          for the Refractive Index of Air, Metrologia, 31: 315-316.
    K.P. Birch and M.J. Downs, 1993, An Updated Edlen Equation for the
          Refractive Index of Air, Metrologia, 30: 155-162.

    Variables used:

    A,B,C,D = constants used to calculate SVP
    A0,A1,A2,B0,B1,C0,C1,DLC,E = constants used to calculate Z
    ALPHA, BETA, GAMMA = constants to calculate enhancement factor of
                          water vapour in air
    CF = correction factor
    CTERM = reference refractivity
    DIST_UC = uncorrected distance (m)
    DIST_CO = corrected distance (m)
    F = enhanvement factor of water vapour in air
    H = humidity (%)
    K0,K1,K2,K3 = constants used to calculate NASX
    LAMDA = wavelength (micrometre)
    MA = molar mass of dry air containing a fraction XC
          of Carbon Dioxide (kg/mol)
    MV = molar mass of water vapour (kg/mol)
    NAS_1 = refractivity of standard air with 450 ppm = (NAS-1)10E8
    NASX_1 = refractivity of standard air with XC ppm = (NASX-1)10E8
    NPROPG_1 = group refractivity = (NPROPgroup-1)10E8
    NPROPP_1 = phase refractivity = (NPROPphase-1)10E8
    NWS_1 = refractivity of water vapour at standard
            conditions = (NWS-1)10E8
    P = pressure (hPa)
    PV = partial water vapour pressure (Pa)
    RHOA = density of dry air (kg per cubic metre)
    RHOASX = density of standard air (kg per cubic metre)
    RHOW = density of pure water vapour (kg per cubic metre)
    RHOWS = density of standard water vapour (kg per cubic metre)
    SIGMA = wavenumber (1/micrometre)
    SVP = saturation vapour pressure of water vapour (Pa)
          in air at temperature (TK)
    TC = temperature (degrees Celcius)
    TK = temperature (Kelvin)
    W0,W1,W2,W3 = constants used to calculate NWS
    XC = carbon dioxide content (ppm)
    Z = compressibility of air
    ZA = compressibility of dry air
    ZWV = compressibility of water vapour
    -----------------------------------------------------------------
"""


def refractivity_constants():
    # PECK & REEDER (1972) AS AMMENDED BY CIDDOR
    # (DRY AIR REFRACTIVITY)
    (K0, K1, K2, K3) = (238.0185, 5792105.0, 57.362, 167917.0)

    # OWENS (1967) WATER VAPOUR REFRACTIVITY
    (W0, W1, W2, W3) = (295.235, 2.6422, -0.032380, 0.004028)

    # ENHANCEMENT FACTOR (DAVIES, 1992)
    (ALPHA, BETA, GAMMA) = (1.00062, 3.14e-8, 5.6e-7)

    # COMPRESSIBILITY (DAVIES 1992)
    (A0, A1, A2, B0, B1, C0, C1, DLC, E) = (
        1.58123e-6, -2.9331e-8, 1.1043e-10, 5.707e-6, -2.051e-8,
        1.9898e-4, -2.376e-6, 1.83e-11, -0.765e-8)

    # CIDDOR 1995 AND DAVIES 1992.  NEW CF ON 28/06/95
    (CF, R, MV) = (1.022, 8.314510, 0.018015)

    # STANDARD CONDITION DRY AIR (PECK & REEDER 1972)
    (TCSTDAIR, TKSTDAIR, PSTDAIR) = (15.0, 288.15, 101325.0)
    (TCSTDWV, TKSTDWV, PSTDWV) = (20.0, 293.15, 1333.0)
    
    return ((K0, K1, K2, K3),
            (W0, W1, W2, W3),
            (ALPHA, BETA, GAMMA),
            (A0, A1, A2, B0, B1, C0, C1, DLC, E),
            (CF, R, MV),
            (TCSTDAIR, TKSTDAIR, PSTDAIR),
            (TCSTDWV, TKSTDWV, PSTDWV))
    

def phase_refractivity(LAMDA, TC, P, PV, XC=420):
    """
    Input
        LAMDA = wavelength (micrometre)
        TC = temperature (degrees Celcius)
        P = pressure (hPa)
        PV = partial water vapour pressure (Pa)
        XC = carbon dioxide content (ppm)
    Return
        NPROPP_1 = phase refractivity = (NPROPphase-1)10E8
    """
    
    ((K0, K1, K2, K3),
    (W0, W1, W2, W3),
    (ALPHA, BETA, GAMMA),
    (A0, A1, A2, B0, B1, C0, C1, DLC, E),
    (CF, R, MV),
    (TCSTDAIR, TKSTDAIR, PSTDAIR),
    (TCSTDWV, TKSTDWV, PSTDWV)) = refractivity_constants()
    
    # CONVERT LAMDA TO SIGMA, THE PRESSURE UNITS TO PASCAL AND CREATE TC AND PP
    SIGMA = 1.0 / LAMDA

    # CONVERSION FROM hPa TO Pa AND C TO K
    P  = P * 100.0
    PV = PV * 100.0
    TK = TC + 273.15

    # CALCULATE REFRACTIVITY OF STANDARD AIR (TC=15, P=101325)
    TEMP1 = SIGMA * SIGMA

    # CIDDOR EQ.(1)
    NAS_1   = (K1 / (K0 - TEMP1)) + (K3 / (K2 - TEMP1))

    # CIDDOR EQ.(2) FOR AMBIENT C02 CONTENT (XC)
    NASX_1 = NAS_1 * (1 + 0.534e-6 * (XC - 450.0))
    STD_AIR = {
            '(Nas-1)10E8': NAS_1,
            '(Nasx-1)10E8':NASX_1}
    
    # USING EQ. 2, CALCULATE REFRACTIVITY OF WATER VAPOUR AT STANDARD CONDITIONS
    # (TC=20, P=1333)
    TEMP2 = TEMP1 * TEMP1
    TEMP3 = TEMP1 * TEMP2

    # FOLLOWING OWENS (1967) AS AMENDED TO FIT BIRCH & DOWNS (1988)
    NWS_1 = CF * (W0 + W1 * TEMP1 + W2 * TEMP2 + W3 * TEMP3)
    STD_WATER_VAPOUR = {
            '(Nws-1)10E8': NWS_1}
    
    # CALCULATE THE DENSITY OF STANDARD AIR (XV=0, TC=15, P=101325)
    # EQ.(2) DAVIS (1992) XC = MOLECULAR FRACTION OF C02
    # XC IN CIDDOR'S FORMULA IS IN PPM, CONVERTED TO A FRACTION
    MA  = (28.9635 + 12.011e-6 * (XC - 400.0)) * 1.0e-3

    # USING CIDDOR EQ.(A1) WITH XV = 0 : DRY AIR COMPRESSIBILITY
    ZA = (1.0 - (PSTDAIR / TKSTDAIR) * (A0 + (A1 * TCSTDAIR) +
         (A2 * TCSTDAIR * TCSTDAIR)) + ((PSTDAIR / TKSTDAIR) *
         (PSTDAIR / TKSTDAIR) * DLC))

    # USING EQ. 3 FOR DRY AIR
    RHOASX = (PSTDAIR * MA) / (ZA * R * TKSTDAIR)
    DENSITY_OF_STD_AIR = {
            'Ma': MA,
            'Z_(DRY AIR)': ZA,
            'RHOasx': RHOASX}
    
    # CALCULATE THE DENSITY OF WATER VAPOUR AT STANDARD CONDITIONS
    # (XV=1, TC=20, P=1333)

    # USING CIDDOR EQUATION A1 WITH XV = 1 AND PARTIAL
    # WATER VAPOUR PRESSURE
    ZWV = (1.0 - (PSTDWV / TKSTDWV) * (A0 + (A1 * TCSTDWV) +
          (A2 * TCSTDWV * TCSTDWV) + (B0 + B1 * TCSTDWV) +
          (C0 + C1 * TCSTDWV)) + ((PSTDWV / TKSTDWV) *
          (PSTDWV / TKSTDWV) * (DLC + E)))

    # USING EQ. 3
    RHOWS = (PSTDWV * MV) / (ZWV * R * TKSTDWV)

    DENSITY_OF_STD_WATER_VAPOUR = {
            'Z_(WATER VAPOUR)': ZWV,
            'RHOws':RHOWS}
    
    # CALCULATE THE DENSITY OF DRY AMBIENT AIR (TC= USER, P=USER)
    F = ALPHA + (BETA * P) + (GAMMA * TC * TC)
    XV = F * PV / P

    # USING CIDDOR EQUATION A1
    Z = (1.0 - (P / TK) * (A0 + (A1 * TC) + (A2 * TC * TC) +
        ((B0 + B1 * TC) * XV) + ((C0 + C1 * TC) * XV * XV)) +
        ((P / TK) * (P / TK) * (DLC + (E * XV * XV))))

    RHOA = (P * MA * (1.0 - XV)) / (Z * R * TK)

    DENSITY_OF_DRY_AMBIENT_AIR = {
            'f': F,
            'Xv': XV,
            'Z_(AMB. AIR)': Z,
            'RHOa':RHOA}
    
    # CALCULATE THE DENSITY OF AMBIENT WATER VAPOUR
    # USING EQ. 3
    RHOW = (P * MV * XV) / (Z * R * TK)
    
    # CALCULATE THE PHASE REFRACTIVITY OF MOIST AIR FROM CIDDOR EQ.(4)
    TEMP1 = RHOA / RHOASX
    TEMP2 = RHOW / RHOWS

    DENSITY_OF_AMBIENT_WATER_VAPOUR = {
        'RHOw' : RHOW,
        'Da' : TEMP1,
        'Dv' : TEMP2}
    
    TEMP1 = TEMP1 * NASX_1
    TEMP2 = TEMP2 * NWS_1
    
    DENSITY_OF_AMBIENT_WATER_VAPOUR['DaNasx'] = TEMP1
    DENSITY_OF_AMBIENT_WATER_VAPOUR['DvNvs'] = TEMP2
        
    NPROPP_1 = TEMP1 + TEMP2
    
    REFRACTIVITY_OF_MOIST_AIR = {
            '(NpropP-1)10E8': NPROPP_1}
    
    return NPROPP_1


def group_refractivity(LAMDA, TC, P, PV, XC=420):
    """
    Input
        LAMDA = wavelength (micrometre)
        TC = temperature (degrees Celcius)
        P = pressure (hPa)
        PV = partial water vapour pressure (Pa)
        XC = carbon dioxide content (ppm)
    Return
        NPROPG_1 = group refractivity = (NPROPgroup-1)10E8
    """
    
    ((K0, K1, K2, K3),
    (W0, W1, W2, W3),
    (ALPHA, BETA, GAMMA),
    (A0, A1, A2, B0, B1, C0, C1, DLC, E),
    (CF, R, MV),
    (TCSTDAIR, TKSTDAIR, PSTDAIR),
    (TCSTDWV, TKSTDWV, PSTDWV)) = refractivity_constants()

    # CONVERT LAMDA TO SIGMA, THE PRESSURE UNITS TO PASCAL AND CREATE TC AND PP
    SIGMA = 1.0 / LAMDA

    # CONVERSION FROM hPa TO Pa AND C TO K
    P  = P * 100.0
    PV = PV * 100.0
    TK = TC + 273.15

    # CALCULATE REFRACTIVITY OF STANDARD AIR (TC=15, P=101325)
    TEMP1 = SIGMA * SIGMA

    # CIDDOR EQ.(9)
    NGAS_1 = (K1 * ((K0 + TEMP1) / ((K0 - TEMP1) * (K0 - TEMP1))) +
              K3 * ((K2 + TEMP1) / ((K2 - TEMP1) * (K2 - TEMP1))))

    # CIDDOR EQ.(2) FOR AMBIENT C02 CONTENT (XC)
    NGASX_1  = NGAS_1 * (1 + 0.534e-6 * (XC - 450.0))
    STD_AIR = {
            '(Ngas-1)10E8': NGAS_1,
            '(Ngasx-1)10E8':NGASX_1}

    # USING EQ. 2, CALCULATE REFRACTIVITY OF WATER VAPOUR AT STANDARD CONDITIONS
    # (TC=20, P=1333)
    TEMP2 = TEMP1 * TEMP1
    TEMP3 = TEMP1 * TEMP2

    # FOLLOWING OWENS (1967) AS AMENDED TO FIT BIRCH & DOWNS (1988)
    NGWS_1 = (CF * (W0 + 3.0 * W1 * TEMP1 + 5.0 * W2 * TEMP2 +
                    7.0 * W3 * TEMP3))
    STD_WATER_VAPOUR = {
            '(Ngws-1)10E8': NGWS_1}
    
    # CALCULATE THE DENSITY OF STANDARD AIR (XV=0, TC=15, P=101325)
    # EQ.(2) DAVIS (1992) XC = MOLECULAR FRACTION OF C02
    # XC IN CIDDOR'S FORMULA IS IN PPM, CONVERTED TO A FRACTION
    MA  = (28.9635 + 12.011e-6 * (XC - 400.0)) * 1.0e-3

    # USING CIDDOR EQ.(A1) WITH XV = 0 : DRY AIR COMPRESSIBILITY
    ZA = (1.0 - (PSTDAIR / TKSTDAIR) * (A0 + (A1 * TCSTDAIR) +
         (A2 * TCSTDAIR * TCSTDAIR)) + ((PSTDAIR / TKSTDAIR) *
         (PSTDAIR / TKSTDAIR) * DLC))

    # USING EQ. 3 FOR DRY AIR
    RHOASX = (PSTDAIR * MA) / (ZA * R * TKSTDAIR)
    DENSITY_OF_STD_AIR = {
            'Ma': MA,
            'Z_(DRY AIR)': ZA,
            'RHOasx': RHOASX}

    # CALCULATE THE DENSITY OF WATER VAPOUR AT STANDARD CONDITIONS
    # (XV=1, TC=20, P=1333)
    
    # USING CIDDOR EQUATION A1 WITH XV = 1 AND PARTIAL
    # WATER VAPOUR PRESSURE
    ZWV = (1.0 - (PSTDWV / TKSTDWV) * (A0 + (A1 * TCSTDWV) +
          (A2 * TCSTDWV * TCSTDWV) + (B0 + B1 * TCSTDWV) +
          (C0 + C1 * TCSTDWV)) + ((PSTDWV / TKSTDWV) *
          (PSTDWV / TKSTDWV) * (DLC + E)))
                                  
    # USING EQ. 3
    RHOWS = (PSTDWV * MV) / (ZWV * R * TKSTDWV)

    DENSITY_OF_STD_WATER_VAPOUR = {
            'Z_(WATER VAPOUR)': ZWV,
            'RHOws':RHOWS}

    # CALCULATE THE DENSITY OF DRY AMBIENT AIR (TC= USER, P=USER)
    F = ALPHA + (BETA * P) + (GAMMA * TC * TC)
    XV = F * PV / P

    # USING CIDDOR EQUATION A1
    Z = (1.0 - (P / TK) * (A0 + (A1 * TC) + (A2 * TC * TC) +
        ((B0 + B1 * TC) * XV) + ((C0 + C1 * TC) * XV * XV)) +
        ((P / TK) * (P / TK) * (DLC + (E * XV * XV))))

    # USING EQ. 3
    RHOA = (P * MA * (1.0 - XV)) / (Z * R * TK)

    DENSITY_OF_DRY_AMBIENT_AIR = {
            'f': F,
            'Xv': XV,
            'Z_(AMB. AIR)': Z,
            'RHOa':RHOA}

    # CALCULATE THE DENSITY OF AMBIENT WATER VAPOUR
    # USING EQ. 3
    RHOW = (P * MV * XV) / (Z * R * TK)

    # CALCULATE THE GROUP REFRACTIVITY OF MOIST AIR FROM CIDDOR EQ.(4)
    TEMP1 = RHOA / RHOASX
    TEMP2 = RHOW / RHOWS

    DENSITY_OF_AMBIENT_WATER_VAPOUR = {
        'RHOw' : RHOW,
        'Da' : TEMP1,
        'Dv' : TEMP2}
    
    TEMP1 = TEMP1 * NGASX_1
    TEMP2 = TEMP2 * NGWS_1

    DENSITY_OF_AMBIENT_WATER_VAPOUR['DaNgasx'] = TEMP1
    DENSITY_OF_AMBIENT_WATER_VAPOUR['DvNgvs'] = TEMP2

    NPROPG_1 = TEMP1 + TEMP2
    
    REFRACTIVITY_OF_MOIST_AIR = {
            '(NpropG-1)10E8': NPROPG_1}
      
    return NPROPG_1
      
    
def humidity2part_water_vapour_press(H, TC):
    """
    Input
        H = humidity (%)
        TC = temperature (degrees Celcius)
    Return
        PV = partial water vapour pressure (Pa)
    """
    
    (A, B, C, D) = (1.2378847e-5, -1.9121316e-2,
                    33.93711047, -6.3431645e3)
    TK = TC + 273.15
    
    # NOW CONVERT THE RELATIVE HUMIDITY TO PATIAL WATER VAPOUR PRESSURE (hPa)
    # SEE EQ.(22) GIACOMO 1982
    SVP = exp(A * TK * TK + B * TK + C + D / TK)
    PV  = (H / 100.0) * SVP
    PV = PV / 100.0

    return PV


#!/usr/bin/env python3

"""
Geoscience Australia - Python Geodesy Package
Angles Module

GeodePy supports Angular Notation in 9 different formats

ABRV - FORMAT (type)
--------------------
rad  - Radians (stored as float)
dec  - Decimal Degrees (stored as float)
deca - Decimal Degrees (via DECAngle class)
hp   - Hewlett Packard (HP) Notation (stored as float)
hpa  - Hewlett Packard (HP) Notation (via HPAngle class)
gon  - Gradians (stored as float)
gona - Gradians (via GONAngle class)
dms  - Degrees, Minutes and Seconds Notation (via DMSAngle class)
ddm  - Degrees and Decimal Minutes Notation (via DDMAngle class)

Conversion between all formats is supported as shown below:

Radians to/from Decimal Degrees via builtin math.radians and math.degrees

Formats as floats to all other types via functions in the form abrv2abrv
e.g. gon2hpa()

DECAngle, HPAngle, GONAngle, DMSAngle and DDMAngle class objects via methods in
the form CLASS.abrv()
e.g. HPAngle(value).dec()

"""

from math import radians


class DECAngle(float):
    """
    Class for working with angles in Decimal Degrees
    Note: GeodePy also supports working with angles in Decimal Degrees as floats
    """

    def __init__(self, dec_angle=0.0):
        """
        :param dec_angle: float Decimal Degrees angle
        """
        super().__init__()
        self.dec_angle = float(dec_angle)

    def __repr__(self):
        if self.dec_angle >= 0:
            return '{DECAngle: +' + str(self.dec_angle) + '}'
        else:  # negative
            return '{DECAngle: ' + str(self.dec_angle) + '}'

    def __add__(self, other):
        try:
            return DECAngle(self.dec() + other.dec())
        except AttributeError:
            raise TypeError('Can only add Angle objects with .dec() method '
                            'together')

    def __radd__(self, other):
        try:
            return DECAngle(other.dec() + self.dec())
        except AttributeError:
            raise TypeError('Can only add Angle objects with .dec() method '
                            'together')

    def __sub__(self, other):
        try:
            return DECAngle(self.dec() - other.dec())
        except AttributeError:
            raise TypeError('Can only subtract Angle objects with .dec() method'
                            ' together')

    def __rsub__(self, other):
        try:
            return DECAngle(other.dec() - self.dec())
        except AttributeError:
            raise TypeError('Can only subtract Angle objects with .dec() method'
                            ' together')

    def __mul__(self, other):
        try:
            return DECAngle(self.dec() * other)
        except TypeError:
            raise TypeError('Multiply only defined between DECAngle Object '
                            'and Int or Float')

    def __rmul__(self, other):
        try:
            return DECAngle(other * self.dec())
        except TypeError:
            raise TypeError('Multiply only defined between DECAngle Object '
                            'and Int or Float')

    def __truediv__(self, other):
        try:
            return DECAngle(self.dec() / other)
        except TypeError:
            raise TypeError('Division only defined between DECAngle Object '
                            'and Int or Float')

    def __abs__(self):
        return DECAngle(abs(self.dec_angle))

    def __neg__(self):
        return DECAngle(-self.dec())

    def __eq__(self, other):
        return self.dec() == other.dec()

    def __ne__(self, other):
        return self.dec() != other.dec()

    def __lt__(self, other):
        return self.dec() < other.dec()

    def __gt__(self, other):
        return self.dec() > other.dec()

    def __int__(self):
        return int(self.dec_angle)

    def __float__(self):
        return float(self.dec_angle)

    def __str__(self):
        return str(self.dec_angle)

    def __round__(self, n=None):
        return DECAngle(round(self.dec_angle, n))

    def rad(self):
        """
        Convert to radians
        :return: radians
        :rtype: float
        """
        return radians(self.dec_angle)

    def dec(self):
        """
        Convert to Decimal Degrees (float)
        :return: Decimal Degrees
        :rtype: float
        """
        return self.dec_angle

    def hp(self):
        """
        Convert to HP Notation
        :return: HP Notation (DDD.MMSSSS)
        :rtype: float
        """
        return dec2hp(self.dec_angle)

    def hpa(self):
        """
        Convert to HP Notation (class)
        :return: HP Notation (DDD.MMSSSS)
        :rtype: HPAngle
        """
        return HPAngle(self.hp())

    def gon(self):
        """
        Convert to Gradians (float)
        :return: Gradians
        :rtype: float
        """
        return dec2gon(self.dec_angle)

    def gona(self):
        """
        Convert to Gradians (class)
        :return: Gradians
        :rtype: GONAngle
        """
        return GONAngle(dec2gon(self.dec_angle))

    def dms(self):
        """
        Convert to Degrees, Minutes, Seconds Object
        :return: Degrees, Minutes, Seconds Object
        :rtype: DMSAngle
        """
        return dec2dms(self.dec_angle)

    def ddm(self):
        """
        Convert to Degrees, Decimal Minutes Object
        :return: Degrees, Decimal Minutes Object
        :rtype: DDMAngle
        """
        return dec2ddm(self.dec_angle)


class HPAngle(object):
    """
    Class for working with angles in Hewlett-Packard (HP) format
    Note: GeodePy also supports working with angles in HP format as floats
    """
    def __init__(self, hp_angle=0.0):
        """
        :param hp_angle: float HP angle
        """
        self.hp_angle = float(hp_angle)
        hp_dec_str = f'{self.hp_angle:.17f}'.split('.')[1]
        if int(hp_dec_str[0]) > 5:
            raise ValueError(f'Invalid HP Notation: 1st decimal place greater '
                             f'than 5: {self.hp_angle}')
        if len(hp_dec_str) > 2:
            if int(hp_dec_str[2]) > 5:
                raise ValueError(
                    f'Invalid HP Notation: 3rd decimal place greater '
                    f'than 5: {self.hp_angle}')

    def __repr__(self):
        if self.hp_angle >= 0:
            return '{HPAngle: +' + str(self.hp_angle) + '}'
        else:  # negative
            return '{HPAngle: ' + str(self.hp_angle) + '}'

    def __add__(self, other):
        try:
            return HPAngle(dec2hp(self.dec() + other.dec()))
        except AttributeError:
            raise TypeError('Can only add Angle objects with .dec() method '
                            'together')

    def __radd__(self, other):
        try:
            return HPAngle(dec2hp(other.dec() + self.dec()))
        except AttributeError:
            raise TypeError('Can only add Angle objects with .dec() method '
                            'together')

    def __sub__(self, other):
        try:
            return HPAngle(dec2hp(self.dec() - other.dec()))
        except AttributeError:
            raise TypeError('Can only subtract Angle objects with .dec() method'
                            ' together')

    def __rsub__(self, other):
        try:
            return HPAngle(dec2hp(other.dec() - self.dec()))
        except AttributeError:
            raise TypeError('Can only subtract Angle objects with .dec() method'
                            ' together')

    def __mul__(self, other):
        try:
            return HPAngle(dec2hp(self.dec() * other))
        except TypeError:
            raise TypeError('Multiply only defined between Angle objects and '
                            'Int or Float')

    def __rmul__(self, other):
        try:
            return HPAngle(dec2hp(other * self.dec()))
        except TypeError:
            raise TypeError('Multiply only defined between Angle objects and '
                            'Int or Float')

    def __truediv__(self, other):
        try:
            return HPAngle(dec2hp(self.dec() / other))
        except TypeError:
            raise TypeError('Division only defined between HPAngle objects '
                            'and Int or Float')

    def __abs__(self):
        return HPAngle(abs(self.hp_angle))

    def __neg__(self):
        return HPAngle(self.hp_angle.__neg__())

    def __eq__(self, other):
        return self.dec() == other.dec()

    def __ne__(self, other):
        return self.dec() != other.dec()

    def __lt__(self, other):
        return self.dec() < other.dec()

    def __gt__(self, other):
        return self.dec() > other.dec()

    def __int__(self):
        return int(self.hp_angle)

    def __float__(self):
        return float(self.hp_angle)

    def __str__(self):
        return str(self.hp_angle)

    def __round__(self, n=None):
        return HPAngle(round(self.hp_angle, n))

    def rad(self):
        """
        Convert to Radians
        :return: Radians
        :rtype: float
        """
        return radians(hp2dec(self.hp_angle))

    def dec(self):
        """
        Convert to Decimal Degrees (float)
        :return: Decimal Degrees
        :rtype: float
        """
        return hp2dec(self.hp_angle)

    def deca(self):
        """
        Convert to Decimal Degrees (class)
        :return: Decimal Degrees
        :rtype: DECAngle
        """
        return DECAngle(self.dec())

    def hp(self):
        """
        Convert to HP Notation (float)
        :return: HP Notation (DDD.MMSSSS)
        :rtype: float
        """
        return float(self.hp_angle)

    def gon(self):
        """
        Convert to Gradians (float)
        :return: Gradians
        :rtype: float
        """
        return hp2gon(self.hp_angle)

    def gona(self):
        """
        Convert to Gradians (class)
        :return: Gradians
        :rtype: GONAngle
        """
        return GONAngle(self.gon())

    def dms(self):
        """
        Convert to Degrees, Minutes, Seconds Object
        :return: Degrees, Minutes, Seconds Object
        :rtype: DMSAngle
        """
        return hp2dms(self.hp_angle)

    def ddm(self):
        """
        Convert to Degrees, Decimal Minutes Object
        :return: Degrees, Decimal Minutes Object
        :rtype: DDMAngle
        """
        return hp2ddm(self.hp_angle)


class GONAngle(object):
    """
    Class for working with angles in Gradians (90 degrees == 100 Gradians)
    Note: GeodePy also supports working with angles in Gradians as floats
    """
    def __init__(self, gon_angle=0.0):
        """
        :param gon_angle: float Gradian angle
        """
        super().__init__()
        self.gon_angle = float(gon_angle)

    def __repr__(self):
        if self.gon_angle >= 0:
            return '{GONAngle: +' + str(self.gon_angle) + '}'
        else:  # negative
            return '{GONAngle: ' + str(self.gon_angle) + '}'

    def __add__(self, other):
        try:
            return GONAngle(dec2gon(self.dec() + other.dec()))
        except AttributeError:
            raise TypeError('Can only add Angle objects with .dec() method '
                            'together')

    def __radd__(self, other):
        try:
            return GONAngle(dec2gon(other.dec() + self.dec()))
        except AttributeError:
            raise TypeError('Can only add Angle objects with .dec() method '
                            'together')

    def __sub__(self, other):
        try:
            return GONAngle(dec2gon(self.dec() - other.dec()))
        except AttributeError:
            raise TypeError('Can only subtract Angle objects with .dec() method'
                            ' together')

    def __rsub__(self, other):
        try:
            return GONAngle(dec2gon(other.dec() - self.dec()))
        except AttributeError:
            raise TypeError('Can only subtract Angle objects with .dec() method'
                            ' together')

    def __mul__(self, other):
        try:
            return GONAngle(dec2gon(self.dec() * other))
        except TypeError:
            raise TypeError('Multiply only defined between Angle objects and '
                            'Int or Float')

    def __rmul__(self, other):
        try:
            return GONAngle(dec2gon(other * self.dec()))
        except TypeError:
            raise TypeError('Multiply only defined between Angle objects and '
                            'Int or Float')

    def __truediv__(self, other):
        try:
            return GONAngle(dec2gon(self.dec() / other))
        except TypeError:
            raise TypeError('Division only defined between HPAngle objects '
                            'and Int or Float')

    def __abs__(self):
        return GONAngle(abs(self.gon_angle))

    def __neg__(self):
        return GONAngle(self.gon_angle.__neg__())

    def __eq__(self, other):
        return self.dec() == other.dec()

    def __ne__(self, other):
        return self.dec() != other.dec()

    def __lt__(self, other):
        return self.dec() < other.dec()

    def __gt__(self, other):
        return self.dec() > other.dec()

    def __int__(self):
        return int(self.gon_angle)

    def __float__(self):
        return float(self.gon_angle)

    def __str__(self):
        return str(self.gon_angle)

    def __round__(self, n=None):
        return GONAngle(round(self.gon_angle, n))

    def rad(self):
        """
        Convert to Radians
        :return: Radians
        :rtype: float
        """
        return radians(gon2dec(self.gon_angle))

    def dec(self):
        """
        Convert to Decimal Degrees (float)
        :return: Decimal Degrees
        :rtype: float
        """
        return gon2dec(self.gon_angle)

    def deca(self):
        """
        Convert to Decimal Degrees (class)
        :return: Decimal Degrees
        :rtype: DECAngle
        """
        return DECAngle(self.dec())

    def hp(self):
        """
        Convert to HP Notation (float)
        :return: HP Notation (DDD.MMSSSS)
        :rtype: float
        """
        return gon2hp(self.gon_angle)

    def hpa(self):
        """
        Convert to HP Notation (class)
        :return: HP Notation (DDD.MMSSSS)
        :rtype: HPAngle
        """
        return HPAngle(gon2hp(self.gon_angle))

    def gon(self):
        """
        Convert to Gradians (float)
        :return: Gradians
        :rtype: float
        """
        return float(self.gon_angle)

    def dms(self):
        """
        Convert to Degrees, Minutes, Seconds Object
        :return: Degrees, Minutes, Seconds Object
        :rtype: DMSAngle
        """
        return gon2dms(self.gon_angle)

    def ddm(self):
        """
        Convert to Degrees, Decimal Minutes Object
        :return: Degrees, Decimal Minutes Object
        :rtype: DDMAngle
        """
        return gon2ddm(self.gon_angle)


class DMSAngle(object):
    """
    Class for working with angles in Degrees, Minutes and Seconds format
    """
    def __init__(self, degree, minute=0, second=0.0, positive=None):
        """
        :param degree: Angle: whole degrees component (floats truncated)
                       Alt: formatted string '±DDD MM SS.SSS'
        :type degree: float | str
        :param minute: Angle: whole minutes component (floats truncated)
        :type minute: float
        :param second: Angle: seconds component (floats preserved)
        :type second: float
        :param positive: Optional. True is positive, False is negative. Evaluated from deg/min/sec where None
        :type positive: bool
        """
        # evaluate sign
        if positive is False or str(degree)[0] == '-':
            self.positive = False
        else:
            self.positive = True

        # check sign provided for minute and second where positive not provided and degree is int == 0
        if degree == 0 and positive is None:
            if minute < 0:
                self.positive = False
            elif second < 0:
                self.positive = False

        # Convert formatted string 'DDD MM SS.SSS' to DMSAngle
        if type(degree) == str:
            str_pts = degree.split()
            degree = int(str_pts[0])
            minute = int(str_pts[1])
            second = float(str_pts[2])

        self.degree = abs(int(degree))
        self.minute = abs(int(minute))
        self.second = abs(second)

    def __repr__(self):
        if self.positive:
            signsymbol = '+'
        else:
            signsymbol = '-'
        return '{DMSAngle: ' + signsymbol + str(self.degree) + 'd ' +\
               str(self.minute) + 'm ' + str(self.second) + 's}'

    def __add__(self, other):
        try:
            return dec2dms(self.dec() + other.dec())
        except AttributeError:
            raise TypeError('Can only add Angle objects with .dec() method '
                            'together')

    def __radd__(self, other):
        try:
            return dec2dms(other.dec() + self.dec())
        except AttributeError:
            raise TypeError('Can only add Angle objects with .dec() method '
                            'together')

    def __sub__(self, other):
        try:
            return dec2dms(self.dec() - other.dec())
        except AttributeError:
            raise TypeError('Can only subtract Angle objects with .dec() method'
                            ' together')

    def __rsub__(self, other):
        try:
            return dec2dms(other.dec() - self.dec())
        except AttributeError:
            raise TypeError('Can only subtract Angle objects with .dec() method'
                            ' together')

    def __mul__(self, other):
        try:
            return dec2dms(self.dec() * other)
        except TypeError:
            raise TypeError('Multiply only defined between DMSAngle Object '
                            'and Int or Float')

    def __rmul__(self, other):
        try:
            return dec2dms(other * self.dec())
        except TypeError:
            raise TypeError('Multiply only defined between DMSAngle Object '
                            'and Int or Float')

    def __truediv__(self, other):
        try:
            return dec2dms(self.dec() / other)
        except TypeError:
            raise TypeError('Division only defined between DMSAngle Object '
                            'and Int or Float')

    def __abs__(self):
        return DMSAngle(self.degree, self.minute, self.second)

    def __neg__(self):
        if self.positive:
            return DMSAngle(-self.degree, -self.minute, -self.second)
        else:  # positive == False
            return DMSAngle(self.degree, self.minute, self.second)

    def __eq__(self, other):
        return self.dec() == other.dec()

    def __ne__(self, other):
        return self.dec() != other.dec()

    def __lt__(self, other):
        return self.dec() < other.dec()

    def __gt__(self, other):
        return self.dec() > other.dec()

    def __str__(self):
        if self.positive:
            return (str(self.degree) + ' ' + str(self.minute) + ' '
                    + str(self.second))
        else:
            return ('-' + str(self.degree) + ' ' + str(self.minute) + ' '
                    + str(self.second))

    def __round__(self, n=None):
        if self.positive:
            return DMSAngle(self.degree, self.minute, round(self.second, n))
        else:
            return -DMSAngle(self.degree, self.minute, round(self.second, n))

    def __mod__(self, other):
        return dec2dms(self.dec() % other)

    def rad(self):
        """
        Convert to Radians
        :return: Radians
        :rtype: float
        """
        return radians(self.dec())

    def dec(self):
        """
        Convert to Decimal Degrees (float)
        :return: Decimal Degrees
        :rtype: float
        """
        if self.positive:
            return self.degree + (self.minute / 60) + (self.second / 3600)
        else:
            return -(self.degree + (self.minute / 60) + (self.second / 3600))

    def deca(self):
        """
        Convert to Decimal Degrees (class)
        :return: Decimal Degrees
        :rtype: DECAngle
        """
        return DECAngle(self.dec())

    def hp(self):
        """
        Convert to HP Notation (float)
        :return: HP Notation (DDD.MMSSSS)
        :rtype: float
        """
        if self.positive:
            return self.degree + (self.minute / 100) + (self.second / 10000)
        else:
            return -(self.degree + (self.minute / 100) + (self.second / 10000))

    def hpa(self):
        """
        Convert to HP Notation (class)
        :return: HP Notation (DDD.MMSSSS)
        :rtype: HPAngle
        """
        return HPAngle(self.hp())

    def gon(self):
        """
        Convert to Gradians (float)
        :return: Gradians
        :rtype: float
        """
        return dec2gon(self.dec())

    def gona(self):
        """
        Convert to Gradians (class)
        :return: Gradians
        :rtype: GONAngle
        """
        return GONAngle(self.gon())

    def ddm(self):
        """
        Convert to Degrees, Decimal Minutes Object
        :return: Degrees, Decimal Minutes Object
        :rtype: DDMAngle
        """
        if self.positive:
            return DDMAngle(self.degree, self.minute + (self.second/60))
        else:
            return -DDMAngle(self.degree, self.minute + (self.second/60))


class DDMAngle(object):
    """
    Class for working with angles in Degrees, Decimal Minutes format
    """
    def __init__(self, degree, minute=0.0, positive=None):
        """
        :param degree: Angle: whole degrees component (floats truncated)
        :type degree: float | str
        :param minute: Angle: minutes component (floats preserved)
        :type minute: float
        :param positive: Optional. True is positive, False is negative. Evaluated from deg/min/sec where None
        :type positive: bool
        """

        # evaluate sign
        if positive is False or str(degree)[0] == '-':
            self.positive = False
        else:
            self.positive = True

        # check sign provided for minute where positive not provided and degree is int == 0
        if degree == 0 and positive is None:
            if minute < 0:
                self.positive = False

        # Convert formatted string 'DDD MM.MMMM' to DDMAngle
        if type(degree) == str:
            str_pts = degree.split(' ')
            degree = int(str_pts[0])
            minute = float(str_pts[1])

        self.degree = abs(int(degree))
        self.minute = abs(minute)

    def __repr__(self):
        if self.positive:
            signsymbol = '+'
        else:
            signsymbol = '-'
        return '{DDMAngle: ' + signsymbol + str(self.degree) + 'd ' + \
               str(self.minute) + 'm}'

    def __add__(self, other):
        try:
            return dec2ddm(self.dec() + other.dec())
        except AttributeError:
            raise TypeError('Can only add Angle objects with .dec() method '
                            'together')

    def __radd__(self, other):
        try:
            return dec2ddm(other.dec() + self.dec())
        except AttributeError:
            raise TypeError('Can only add Angle objects with .dec() method '
                            'together')

    def __sub__(self, other):
        try:
            return dec2ddm(self.dec() - other.dec())
        except AttributeError:
            raise TypeError('Can only subtract Angle objects with .dec() method'
                            ' together')

    def __rsub__(self, other):
        try:
            return dec2ddm(other.dec() - self.dec())
        except AttributeError:
            raise TypeError('Can only subtract Angle objects with .dec() method'
                            ' together')

    def __mul__(self, other):
        try:
            return dec2ddm(self.dec() * other)
        except TypeError:
            raise TypeError('Multiply only defined between DMSAngle Object '
                            'and Int or Float')

    def __rmul__(self, other):
        try:
            return dec2ddm(other * self.dec())
        except TypeError:
            raise TypeError('Multiply only defined between DMSAngle Object '
                            'and Int or Float')

    def __truediv__(self, other):
        try:
            return dec2ddm(self.dec() / other)
        except TypeError:
            raise TypeError('Division only defined between DMSAngle Object '
                            'and Int or Float')

    def __abs__(self):
        return DDMAngle(self.degree, self.minute)

    def __neg__(self):
        if self.positive:
            return DDMAngle(-self.degree, -self.minute)
        else:  # sign == -1
            return DDMAngle(self.degree, self.minute)

    def __eq__(self, other):
        return self.dec() == other.dec()

    def __ne__(self, other):
        return self.dec() != other.dec()

    def __lt__(self, other):
        return self.dec() < other.dec()

    def __gt__(self, other):
        return self.dec() > other.dec()

    def __str__(self):
        if self.positive:
            return str(self.degree) + ' ' + str(self.minute)
        else:
            return '-' + str(self.degree) + ' ' + str(self.minute)

    def __round__(self, n=None):
        if self.positive:
            return DDMAngle(self.degree, round(self.minute, n))
        else:
            return DDMAngle(-self.degree, -round(self.minute, n))

    def __mod__(self, other):
        return dec2ddm(self.dec() % other)


    def rad(self):
        """
        Convert to Radians
        :return: Radians
        :rtype: float
        """
        return radians(self.dec())

    def dec(self):
        """
        Convert to Decimal Degrees (float)
        :return: Decimal Degrees
        :rtype: float
        """
        if self.positive:
            return self.degree + (self.minute / 60)
        else:
            return -(self.degree + (self.minute / 60))

    def deca(self):
        """
        Convert to Decimal Degrees (class)
        :return: Decimal Degrees
        :rtype: DECAngle
        """
        return DECAngle(self.dec())

    def hp(self):
        """
        Convert to HP Notation (float)
        :return: HP Notation (DDD.MMSSSS)
        :rtype: float
        """
        minute_int, second = divmod(self.minute, 1)
        if self.positive:
            return self.degree + (minute_int / 100) + (second * 0.006)
        else:
            return -(self.degree + (minute_int / 100) + (second * 0.006))

    def hpa(self):
        """
        Convert to HP Notation (class)
        :return: HP Notation (DDD.MMSSSS)
        :rtype: HPAngle
        """
        return HPAngle(self.hp())

    def gon(self):
        """
        Convert to Gradians (float)
        :return: Gradians
        :rtype: float
        """
        return dec2gon(self.dec())

    def gona(self):
        """
        Convert to Gradians (class)
        :return: Gradians
        :rtype: GONAngle
        """
        return GONAngle(self.gon())

    def dms(self):
        """
        Convert to Degrees, Minutes, Seconds Object
        :return: Degrees, Minutes, Seconds Object
        :rtype: DMSAngle
        """
        minute_int, second = divmod(self.minute, 1)
        if self.positive:
            return DMSAngle(self.degree, int(minute_int), second * 60)
        else:
            return -DMSAngle(self.degree, int(minute_int), second * 60)


# Functions converting from Decimal Degrees (float) to other formats

def dec2hp(dec):
    """
    Converts Decimal Degrees to HP Notation (float)
    :param dec: Decimal Degrees
    :type dec: float
    :return: HP Notation (DDD.MMSSSS)
    :rtype: float
    """      
    minute, second = divmod(abs(dec) * 3600, 60)
    degree, minute = divmod(minute, 60)

    # floating point precision is 13 places for the variable 'dec' where values
    # are between 256 and 512 degrees. Precision improves for smaller angles.
    # In calculating the variable 'second' the precision is degraded by a factor of 3600 
    # Therefore 'second' should be rounded to 9 DP and tested for carry.
    if round(second, 9) == 60:
        second = 0
        minute += 1
    
    # to avoid precision issues with floating point operations
    # a string will be built to represent a sexagesimal number and then converted to float
    degree = f'{int(degree)}'
    minute = f'{int(minute):02}'
    second = f'{second:012.9f}'.rstrip('0').replace('.', '')
    
    hp_string = f'{degree}.{minute}{second}'
    hp = float(hp_string)
     
    return hp if dec >= 0 else -hp


def dec2hpa(dec):
    """
    Converts Decimal Degrees to HP Angle Object
    :param dec: Decimal Degrees
    :type dec: float
    :return: HP Angle Object (DDD.MMSSSS)
    :rtype: HPAngle
    """
    return HPAngle(dec2hp(dec))


def dec2gon(dec):
    """
    Converts Decimal Degrees to Gradians
    :param dec: Decimal Degrees
    :type dec: float
    :return: Gradians
    :rtype: float
    """
    return 10/9 * dec


def dec2gona(dec):
    """
    Converts Decimal Degrees to Gradians (class)
    :param dec: Decimal Degrees
    :type dec: float
    :return: Gradians
    :rtype: GONAngle
    """
    return GONAngle(dec2gon(dec))


def dec2dms(dec):
    """
    Converts Decimal Degrees to Degrees, Minutes, Seconds Object
    :param dec: Decimal Degrees
    :type dec: float
    :return: Degrees, Minutes, Seconds Object
    :rtype: DMSAngle
    """
    minute, second = divmod(abs(dec) * 3600, 60)
    degree, minute = divmod(minute, 60)
    return (DMSAngle(degree, minute, second, positive=True) if dec >= 0
            else DMSAngle(degree, minute, second, positive=False))


def dec2ddm(dec):
    """
    Converts Decimal Degrees to Degrees, Decimal Minutes Object
    :param dec: Decimal Degrees
    :type dec: float
    :return: Degrees, Decimal Minutes Object
    :rtype: DDMAngle
    """
    minute, second = divmod(abs(dec) * 3600, 60)
    degree, minute = divmod(minute, 60)
    minute = minute + (second / 60)
    return DDMAngle(degree, minute, positive=True) if dec >= 0 else DDMAngle(degree, minute, positive=False)


# Functions converting from Hewlett-Packard (HP) format to other formats

def hp2dec(hp):
    """
    Converts HP Notation to Decimal Degrees
    :param hp: HP Notation (DDD.MMSSSS)
    :type hp: float
    :return: Decimal Degrees
    :rtype: float
    """
    # Check if 1st and 3rd decimal place greater than 5 (invalid HP Notation)
    hp = float(hp)
    hp_deg_str, hp_mmss_str = f'{hp:.13f}'.split('.')
    if int(hp_mmss_str[0]) > 5:
        raise ValueError(f'Invalid HP Notation: 1st decimal place greater '
                         f'than 5: {hp}')
    if len(hp_mmss_str) > 2:
        if int(hp_mmss_str[2]) > 5:
            raise ValueError(f'Invalid HP Notation: 3rd decimal place greater '
                             f'than 5: {hp}')
    # parse string to avoid precision problems with floating point ops and base 10 numbers
    deg = abs(int(hp_deg_str))
    min = int(hp_mmss_str[:2])
    sec = float(hp_mmss_str[2:4] + '.' + hp_mmss_str[4:])
    dec = sec / 3600 + min / 60 + deg

    return dec if hp >= 0 else -dec


def hp2deca(hp):
    """
    Converts HP Notation to DECAngle Object
    :param hp: HP Notation (DDD.MMSSSS)
    :type hp: float
    :return: Decimal Degrees Object
    :rtype: DECAngle
    """
    return DECAngle(hp2dec(hp))


def hp2rad(hp):
    """
    Converts HP Notation to radians
    :param hp: HP Notation (DDD.MMSSSS)
    :type hp: float
    :return: radians
    :rtype: float
    """
    return radians(hp2dec(hp))


def hp2gon(hp):
    """
    Converts HP Notation to Gradians
    :param hp: HP Notation (DDD.MMSSSS)
    :type hp: float
    :return: Gradians
    :rtype: float
    """
    return dec2gon(hp2dec(hp))


def hp2gona(hp):
    """
    Converts HP Notation to Gradians (class)
    :param hp: HP Notation (DDD.MMSSSS)
    :type hp: float
    :return: Gradians
    :rtype: GONAngle
    """
    return GONAngle(hp2gon(hp))


def hp2dms(hp):
    """
    Converts HP Notation to Degrees, Minutes, Seconds Object
    :param hp: HP Notation (DDD.MMSSSS)
    :type hp: float
    :return: Degrees, Minutes, Seconds Object
    :rtype: DMSAngle
    """
    degmin, second = divmod(abs(hp) * 1000, 10)
    degree, minute = divmod(degmin, 100)
    return (DMSAngle(degree, minute, second * 10, positive=True) if hp >= 0
            else DMSAngle(degree, minute, second * 10, positive=False))


def hp2ddm(hp):
    """
    Converts HP Notation to Degrees, Decimal Minutes Object
    :param hp: HP Notation (DDD.MMSSSS)
    :type hp: float
    :return: Degrees, Decimal Minutes Object
    :rtype: DDMAngle
    """
    degmin, second = divmod(abs(hp) * 1000, 10)
    degree, minute = divmod(degmin, 100)
    minute = minute + (second / 6)
    return DDMAngle(degree, minute, positive=True) if hp >= 0 else DDMAngle(degree, minute, positive=False)


# Functions converting from Gradians format to other formats

def gon2dec(gon):
    """
    Converts Gradians to Decimal Degrees
    :param gon: Gradians
    :type gon: float
    :return: Decimal Degrees
    :rtype: float
    """
    return 9/10 * gon


def gon2deca(gon):
    """
    Converts Gradians to DECAngle Object
    :param gon: Gradians
    :type gon: float
    :return: Decimal Degrees Object
    :rtype: DECAngle
    """
    return DECAngle(gon2dec(gon))


def gon2hp(gon):
    """
    Converts Gradians to HP Notation (float)
    :param gon: Gradians
    :type gon: float
    :return: HP Notation (DDD.MMSSSS)
    :rtype: float
    """
    return dec2hp(gon2dec(gon))


def gon2hpa(gon):
    """
    Converts Gradians to HP Angle Object
    :param gon: Gradians
    :type gon: float
    :return: HP Angle Object (DDD.MMSSSS)
    :rtype: HPAngle
    """
    return HPAngle(gon2hp(gon))


def gon2rad(gon):
    """
    Converts Gradians to radians
    :param gon: Gradians
    :type gon: float
    :return: Radians
    :rtype: float
    """
    return radians(gon2dec(gon))


def gon2dms(gon):
    """
    Converts Gradians to Degrees, Minutes, Seconds Object
    :param gon: Gradians
    :type gon: float
    :return: Degrees, Minutes, Seconds Object
    :rtype: DMSAngle
    """
    return dec2dms(gon2dec(gon))


def gon2ddm(gon):
    """
    Converts Gradians to Degrees, Decimal Minutes Object
    :param gon: Gradians
    :type gon: float
    :return: Degrees, Decimal Minutes Object
    :rtype: DDMAngle
    """
    return dec2ddm(gon2dec(gon))


# Miscellaneous other useful functions

def dd2sec(dd):
    """
    Converts angle in decimal degrees to angle in seconds
    :param dd: Decimal Degrees
    :return: Seconds
    """
    minute, second = divmod(abs(dd) * 3600, 60)
    degree, minute = divmod(minute, 60)
    sec = (degree * 3600) + (minute * 60) + second
    return sec if dd >= 0 else -sec


def dec2hp_v(dec):
    minute, second = divmod(abs(dec) * 3600, 60)
    degree, minute = divmod(minute, 60)
    hp = degree + (minute / 100) + (second / 10000)
    hp[dec <= 0] = -hp[dec <= 0]
    return hp


def hp2dec_v(hp):
    degmin, second = divmod(abs(hp) * 1000, 10)
    degree, minute = divmod(degmin, 100)
    dec = degree + (minute / 60) + (second / 360)
    dec[hp <= 0] = -dec[hp <= 0]
    return dec


def angular_typecheck(angle):
    # Converts Angle Objects to Decimal Degrees (float) for computations
    supported_types = [DMSAngle, DDMAngle, DECAngle, HPAngle, GONAngle]
    if type(angle) in supported_types:
        return angle.dec()
    else:
        return float(angle)



#!/usr/bin/env python3

"""
Geoscience Australia - Python Geodesy Package
Transform Module

Ref1
http://www.icsm.gov.au/sites/default/files/GDA2020TechnicalManualV1.1.1.pdf

Ref2
http://www.mygeodesy.id.au/documents/Karney-Krueger%20equations.pdf
"""

import datetime
from math import radians
import numpy as np
from geodepy.constants import (Transformation, TransformationSD,
                               atrf2014_to_gda2020, gda94_to_gda2020)
from geodepy.statistics import vcv_local2cart, vcv_cart2local
from geodepy.convert import (hp2dec, geo2grid,
                             grid2geo, xyz2llh, llh2xyz)
from geodepy.ntv2reader import NTv2Grid, interpolate_ntv2


def conform7(x, y, z, trans, vcv=None):
    """
    Performs a Helmert 7 Parameter Conformal Transformation using Cartesian point co-ordinates
    and a predefined transformation object.
    :param x: Cartesian X (m)
    :param y: Cartesian Y (m)
    :param z: Cartesian Z (m)
    :param trans: Transformation Object (note: this function ignores all time-dependent variables)
    :param vcv: Optional 3*3 numpy array in Cartesian units to propagate tf uncertainty
    :return: Transformed X, Y, Z Cartesian Co-ordinates, vcv matrix
    """
    if type(trans) != Transformation:
        raise ValueError('trans must be a Transformation Object')
    # Create XYZ Vector
    xyz_before = np.array([[x],
                           [y],
                           [z]])
    # Convert Units for Transformation Parameters
    scale = 1 + trans.sc / 1000000
    rx = radians(hp2dec(trans.rx / 10000))
    ry = radians(hp2dec(trans.ry / 10000))
    rz = radians(hp2dec(trans.rz / 10000))
    # Create Translation Vector
    translation = np.array([[trans.tx],
                            [trans.ty],
                            [trans.tz]])
    # Create Rotation Matrix
    rotation = np.array([[1., rz, -ry],
                         [-rz, 1., rx],
                         [ry, -rx, 1.]])

    rot_xyz = rotation @ xyz_before

    # Conformal Transform Eq
    xyz_after = translation + scale * rot_xyz
    # Convert Vector to Separate Variables
    xtrans = float(xyz_after[0][0])
    ytrans = float(xyz_after[1][0])
    ztrans = float(xyz_after[2][0])

    # Transformation uncertainty propagation
    # Adapted from Harvey B.R. (1998) Practical least squares and statistics for surveyors,
    # Monograph 13 Section 8.7.2, p.274
    if (type(trans.tf_sd) == TransformationSD) and (vcv is not None):
        # Q matrix:
        q_mat = np.zeros((10, 10))
        # xyz_before vcv
        for i in range(3):
            for j in range(3):
                q_mat[i, j] = vcv[i, j]

        # transformation variances
        q_mat[3, 3] = (trans.tf_sd.sd_sc / 1000000)**2
        q_mat[4, 4] = radians(trans.tf_sd.sd_rx/3600)**2
        q_mat[5, 5] = radians(trans.tf_sd.sd_ry/3600)**2
        q_mat[6, 6] = radians(trans.tf_sd.sd_rz/3600)**2
        q_mat[7, 7] = trans.tf_sd.sd_tx**2
        q_mat[8, 8] = trans.tf_sd.sd_ty**2
        q_mat[9, 9] = trans.tf_sd.sd_tz**2

        # Jacobian matrix:
        j_mat = np.zeros((3, 10))

        # scaled rotations
        j_mat[0, 0] = scale
        j_mat[0, 1] = scale * rz
        j_mat[0, 2] = -scale * ry
        j_mat[1, 0] = -j_mat[0, 1]
        j_mat[1, 1] = scale
        j_mat[1, 2] = scale * rx
        j_mat[2, 0] = -j_mat[0, 2]
        j_mat[2, 1] = -j_mat[1, 2]
        j_mat[2, 2] = scale

        # XYZ rotated
        j_mat[0, 3] = rot_xyz[0]
        j_mat[1, 3] = rot_xyz[1]
        j_mat[2, 3] = rot_xyz[2]

        # scaled XYZ
        j_mat[0, 5] = -scale * xyz_before[2]
        j_mat[0, 6] = scale * xyz_before[1]
        j_mat[1, 4] = scale * xyz_before[2]
        j_mat[1, 6] = -scale * xyz_before[0]
        j_mat[2, 4] = -scale * xyz_before[1]
        j_mat[2, 5] = scale * xyz_before[0]

        # Identity
        j_mat[0, 7] = 1
        j_mat[1, 8] = 1
        j_mat[2, 9] = 1

        # multiply J Q J_trans
        vcv_after = j_mat @ q_mat @ j_mat.transpose()

        return xtrans, ytrans, ztrans, vcv_after

    else:
        return xtrans, ytrans, ztrans, None


def conform14(x, y, z, to_epoch, trans, vcv=None):
    """
    Performs a Helmert 14 Parameter Conformal Transformation using Cartesian point co-ordinates
    and a predefined transformation object. The transformation parameters are projected from
    the transformation objects reference epoch to a specified epoch.
    :param x: Cartesian X (m)
    :param y: Cartesian Y (m)
    :param z: Cartesian Z (m)
    :param to_epoch: Epoch co-ordinate transformation is performed at (datetime.date Object)
    :param trans: Transformation Object
    :param vcv: Optional 3*3 numpy array in Cartesian units to propagate tf uncertainty
    :return: Cartesian X, Y, Z co-ordinates and vcv matrix transformed using Transformation parameters at desired epoch
    """
    if type(trans) != Transformation:
        raise ValueError('trans must be a Transformation Object')
    if type(to_epoch) != datetime.date:
        raise ValueError('to_epoch must be a datetime.date Object')
    # Calculate 7 Parameters from 14 Parameter Transformation Object
    timetrans = trans + to_epoch

    # Perform Transformation
    xtrans, ytrans, ztrans, trans_vcv = conform7(x, y, z, timetrans, vcv=vcv)
    return xtrans, ytrans, ztrans, trans_vcv


def transform_mga94_to_mga2020(zone, east, north, ell_ht=False, vcv=None):
    """
    Performs conformal transformation of Map Grid of Australia 1994 to Map Grid of Australia 2020 Coordinates
    using the GDA2020 Tech Manual v1.2 7 parameter similarity transformation parameters
    :param zone: Zone Number - 1 to 60
    :param east: Easting (m, within 3330km of Central Meridian)
    :param north: Northing (m, 0 to 10,000,000m)
    :param ell_ht: Ellipsoid Height (m) (optional)
    :param vcv: Optional 3*3 numpy array in local enu units to propagate tf uncertainty
    :return: MGA2020 Zone, Easting, Northing, Ellipsoid Height (if none provided, returns 0), and vcv matrix
    """
    lat, lon, psf, gridconv = grid2geo(zone, east, north)
    if ell_ht is False:
        ell_ht_in = 0
    else:
        ell_ht_in = ell_ht
    if vcv is not None:
        vcv = vcv_local2cart(vcv, lat, lon)
    x94, y94, z94 = llh2xyz(lat, lon, ell_ht_in)
    x20, y20, z20, vcv20 = conform7(x94, y94, z94, gda94_to_gda2020, vcv)
    lat, lon, ell_ht_out = xyz2llh(x20, y20, z20)
    if vcv20 is not None:
        vcv20 = vcv_cart2local(vcv20, lat, lon)
    if ell_ht is False:
        ell_ht_out = 0
    hemisphere, zone20, east20, north20, psf, gridconv = geo2grid(lat, lon)
    return zone20, east20, north20, round(ell_ht_out, 4), vcv20


def transform_mga2020_to_mga94(zone, east, north, ell_ht=False, vcv=None):
    """
    Performs conformal transformation of Map Grid of Australia 2020 to Map Grid of Australia 1994 Coordinates
    using the reverse form of the GDA2020 Tech Manual v1.2 7 parameter similarity transformation parameters
    :param zone: Zone Number - 1 to 60
    :param east: Easting (m, within 3330km of Central Meridian)
    :param north: Northing (m, 0 to 10,000,000m)
    :param ell_ht: Ellipsoid Height (m) (optional)
    :param vcv: Optional 3*3 numpy array in local enu units to propagate tf uncertainty
    :return: MGA1994 Zone, Easting, Northing, Ellipsoid Height (if none provided, returns 0), and vcv matrix
    """
    lat, lon, psf, gridconv = grid2geo(zone, east, north)
    if ell_ht is False:
        ell_ht_in = 0
    else:
        ell_ht_in = ell_ht
    if vcv is not None:
        vcv = vcv_local2cart(vcv, lat, lon)
    x94, y94, z94 = llh2xyz(lat, lon, ell_ht_in)
    x20, y20, z20, vcv94 = conform7(x94, y94, z94, -gda94_to_gda2020, vcv=vcv)
    lat, lon, ell_ht_out = xyz2llh(x20, y20, z20)
    if vcv94 is not None:
        vcv94 = vcv_cart2local(vcv94, lat, lon)
    if ell_ht is False:
        ell_ht_out = 0
    hemisphere, zone20, east20, north20, psf, gridconv = geo2grid(lat, lon)
    return zone20, east20, north20, round(ell_ht_out, 4), vcv94


def transform_atrf2014_to_gda2020(x, y, z, epoch_from, vcv=None):
    """
    Transforms Cartesian (x, y, z) Coordinates in terms of the Australian Terrestrial Reference Frame (ATRF) at
    a specified epoch to coordinates in terms of Geocentric Datum of Australia 2020 (GDA2020 - reference epoch 2020.0)
    :param x: ATRF Cartesian X Coordinate (m)
    :param y: ATRF Cartesian Y Coordinate (m)
    :param z: ATRF Cartesian Z Coordinate (m)
    :param epoch_from: ATRF Coordinate Epoch (datetime.date Object)
    :param vcv: Optional 3*3 numpy array in Cartesian units to propagate tf uncertainty
    :return: Cartesian X, Y, Z Coordinates and vcv matrix in terms of GDA2020
    """
    return conform14(x, y, z, epoch_from, atrf2014_to_gda2020, vcv=vcv)


def transform_gda2020_to_atrf2014(x, y, z, epoch_to, vcv=None):
    """
    Transforms Cartesian (x, y, z) Coordinates in terms of Geocentric Datum of Australia 2020
    (GDA2020 - reference epoch 2020.0) to coordinates in terms of the Australian Terrestrial Reference Frame (ATRF) at
    a specified epoch
    :param x: GDA2020 Cartesian X Coordinate (m)
    :param y: GDA2020 Cartesian Y Coordinate (m)
    :param z: GDA2020 Cartesian Z Coordinate (m)
    :param epoch_to: ATRF Coordinate Epoch (datetime.date Object)
    :param vcv: Optional 3*3 numpy array in Cartesian units to propagate tf uncertainty
    :return: Cartesian X, Y, Z Coordinates and vcv matrix in terms of ATRF at the specified Epoch
    """
    return conform14(x, y, z, epoch_to, -atrf2014_to_gda2020, vcv=vcv)


def ntv2_2d(ntv2_grid, lat, lon, forward_tf=True, method='bicubic'):
    """
    Performs a 2D transformation based on ntv2 grid shifts.
    :param ntv2_grid: Ntv2Grid object (create with read_ntv2_file() function in geodepy.ntv2reader module)
    :param lat: latitude in decimal degrees
    :param lon: longitude in decimal degrees
    :param forward_tf: True/False:
                       - True applies the shifts in the direction given in the grid.
                       - False applies the shifts in the opposite direction of the grid
    :param method: Interpolation strategy - either 'bicubic' or 'bilinear'
    :return: Transformed latitude and longitude
    """

    # validate input data
    if not isinstance(ntv2_grid, NTv2Grid):
        raise TypeError('ntv2_grid must be Ntv2Grid object')
    if method != 'bicubic' and method != 'bilinear':
        raise ValueError(f'interpolation strategy "{method}" not supported')

    # interrogate grid
    shifts = interpolate_ntv2(ntv2_grid, lat, lon, method=method)

    # null results are outside of grid extents.
    if shifts[0] is None:
        raise ValueError('Coordinate outside of grid extents')

    if forward_tf:
        tf_lat = lat + shifts[0] / 3600
        tf_lon = lon - shifts[1] / 3600
    else:
        tf_lat = lat - shifts[0] / 3600
        tf_lon = lon + shifts[1] / 3600

    return tf_lat, tf_lon



from setuptools import setup

setup(name='geodepy',
      version='0.6.0',
      description='GA Geodesy Package',
      long_description='A toolkit for Geodesy and Surveying in Python',
      url='https://github.com/GeoscienceAustralia/GeodePy',
      author='Geoscience Australia',
      author_email='geodesy@ga.gov.au',
      license='Apache License 2.0',
      packages=['geodepy'],
      install_requires=['numpy', 'scipy'])



#___________________________________________________________________________#
# Some notes:
# Written by Jack McCubbine of Geoscience Australia, date: 08/11/2019
# This code contains functions to handle tranformations between GPS and 
# AWVS/AHD and Vice Versa 
# Gridded data used for the varisous reference surfaces are geotiff files
# These allow direct access remotely using "gdal"
#___________________________________________________________________________#
# Import dependencies 
import geodepy.constants as cons
import geodepy.geodesy as gg
from osgeo import gdal
import numpy as np
from scipy.interpolate import griddata
import math as m
#___________________________________________________________________________#
# Interpolation functions
def interp_file(Lat,Long,file):
    # Import the DOVPM file
    f = gdal.Open(file)
    # load band (akin to a variable in dataset)
    band = f.GetRasterBand(1)   
    # get the pixel width, height, etc.
    transform = f.GetGeoTransform()
    # Grid resolution (known)
    res=transform[1]
    # convert lat,lon to row,col
    column = (Long - transform[0]) / transform[1]
    row = (Lat - transform[3]) / transform[5]
    # get pixel values surrounding data point
    Surrounding_data=(band.ReadAsArray(np.floor(column-2), np.floor(row-2), 5, 5))
    # convert row,col back to north,east
    Long_c = transform[0] + np.floor(column) * res
    Lat_c = transform[3] - np.floor(row) * res
    # set up matrices for interpolation
    count=-1
    pos=np.zeros((25,2))
    Surrounding_data_v=np.zeros((25,1))
    for k in range(-2,3):
        for j in range(-2,3):
            count=count+1
            pos[count]=(Long_c+j*res,Lat_c-k*res)
            Surrounding_data_v[count]=Surrounding_data[k+2,j+2]          
    interp_val=griddata(pos,Surrounding_data_v,(Long,Lat),method='cubic')
    return interp_val

#___________________________________________________________________________#
# Functions to handle the conversions from one height to another
def GPS_to_AVWS(Lat,Long,GPS_H):
    zeta=interp_file(Lat, Long, cons.file_AVWS) # AVWS file
    zeta_std=interp_file(Lat, Long, cons.file_AVWS_STD) # AVWS STD file
    NORMAL_H=GPS_H-zeta
    return [NORMAL_H,zeta_std]

def AVWS_to_GPS(Lat,Long,AVWS_H):
    zeta=interp_file(Lat, Long, cons.file_AVWS) # AVWS file
    zeta_std=interp_file(Lat, Long, cons.file_AVWS_STD) # AVWS STD file
    GPS_H=AVWS_H+zeta
    return [GPS_H,zeta_std]

def AHD_to_AVWS(Lat,Long,AHD_H):
    # Convert to GPS
    GPS_H=AHD_H+interp_file(Lat, Long, cons.file_AG2020) # AUSGEOID2020 file
    # Convert to AVWS
    Normal_H=GPS_H-interp_file(Lat, Long, cons.file_AVWS) # AVWS file
    return [Normal_H]

def GPS_to_AHD(Lat,Long,GPS_H):
    N=interp_file(Lat, Long, cons.file_AG2020) # AUSGEOID2020 file
    N_std=interp_file(Lat, Long, cons.file_AG2020_STD) # AUSGEOID2020 STD file
    AHD_H=GPS_H-N
    return [AHD_H,N_std]

def AHD_to_GPS(Lat,Long,AHD_H):
    N=interp_file(Lat, Long, cons.file_AG2020) # AUSGEOID2020 file
    N_std=interp_file(Lat, Long, cons.file_AG2020_STD) # AUSGEOID2020 STD file
    GPS_H=AHD_H+N
    return [GPS_H,N_std]

def AVWS_to_AHD(Lat,Long,Normal_H):   
    # Convert to GPS    
    GPS_H=Normal_H+interp_file(Lat, Long, cons.file_AVWS) # AVWS file
    # Convert to AHD
    AHD_H=GPS_H-interp_file(Lat, Long, cons.file_AG2020) # AUSGEOID2020 file
    return [AHD_H]

def DOV(Lat,Long):   
    # Convert to GPS    
    DOV_PM=interp_file(Lat, Long, cons.file_DOV_PM) # AVWS file
    # Convert to AHD
    DOV_PV=interp_file(Lat, Long, cons.file_DOV_PV) # AUSGEOID2020 file
    return [DOV_PM,DOV_PV]

def GPS_to_AUSGeoid98(Lat,Long,GPS_H):
    N=interp_file(Lat,Long,cons.file_AG98) # AUSGEOID98 file
    AHD_H=GPS_H-N
    return [AHD_H]

def AUSGeoid98_to_GPS(Lat,Long,AHD_H):
    N=interp_file(Lat,Long,cons.file_AG98) # AUSGEOID98 file
    GPS_H=AHD_H+N
    return [GPS_H]

def GPS_to_AUSGeoid09(Lat,Long,GPS_H):
    N=interp_file(Lat,Long,cons.file_AG09) # AUSGEOID09 file
    AHD_H=GPS_H-N
    return [AHD_H]

def AUSGeoid09_to_GPS(Lat,Long,AHD_H):
    N=interp_file(Lat,Long,cons.file_AG09) # AUSGEOID09 file
    GPS_H=AHD_H+N
    return [GPS_H]

def DOV_09(Lat,Long):   
    # Interp PM
    DOV_PM=interp_file(Lat,Long,cons.file_AG09_DOV_PM) # AGQG09 DOV file
    # Interp PV
    DOV_PV=interp_file(Lat,Long,cons.file_AG09_DOV_PV) # AGQG09 DOV file
    return [DOV_PM,DOV_PV]

def DOV_98(Lat,Long):   
    # Interp PM  
    DOV_PM=interp_file(Lat,Long,cons.file_AG98_DOV_PM) # AGQG98 DOV file
    # Interp PV
    DOV_PV=interp_file(Lat,Long,cons.file_AG98_DOV_PV) # AGQG98 DOV file
    return [DOV_PM,DOV_PV]

def mean_normal_grav(Lat,h):
    # GRS 80 constants
    a=6378137
    b=6356752.3141
    omega=7292115*(10**-11)
    e2=0.00669438002290
    GM=3986005*10**8
    k=0.001931851353
    # GRS80 normal gravity
    EllGrav=(10**5)*9.7803267715*(1+k*(np.sin(Lat*np.pi/180)**2))/np.sqrt(1-e2*(np.sin(Lat*np.pi/180)**2))
    FA=-((2*(EllGrav/a)*(1+(a-b)/a + omega**2*a**2*b/GM - 2*(a-b)/a*(np.sin(Lat*np.pi/180)**2))*(h**2)/2-3*(EllGrav/a**2)*(h**3)/3)/h)
    mean_normal_g=(EllGrav+FA)*(10**-5)
    return mean_normal_g

def normal_grav(Lat,h):
    # GRS 80 constants
    a=6378137
    b=6356752.3141
    omega=7292115*(10**-11)
    e2=0.00669438002290
    GM=3986005*10**8
    k=0.001931851353
    # GRS80 normal gravity
    EllGrav=(10**5)*9.7803267715*(1+k*(np.sin(Lat*np.pi/180)**2))/np.sqrt(1-e2*(np.sin(Lat*np.pi/180)**2))
    FA=-(2*EllGrav*h/a)*(1+(a-b)/a+omega**2*a**2*b/GM-2*(a-b)/a*(np.sin(Lat*np.pi/180)**2))+3*(EllGrav*h**2)/(a**2)
    normal_g=(EllGrav+FA)*(10**-5)
    return normal_g

def mean_surface_grav(Lat_A,Long_A,H_A,Lat_B,Long_B,H_B):
    Surf_Grav_A=interp_grav(Lat_A,Long_A)*(10**-5)+normal_grav(Lat_A,H_A)+0.0419*2.67*H_A*(10**-5)
    Surf_Grav_B=interp_grav(Lat_B,Long_B)*(10**-5)+normal_grav(Lat_B,H_B)+0.0419*2.67*H_B*(10**-5)
    mean_g=(Surf_Grav_A+Surf_Grav_B)/2
    return mean_g

def interp_grav(Lat,Long):
    # Grid resolution (known)
    res=1.0/60
    # open geotiff file
    f = gdal.Open(cons.file_GRAV_BA)
    # load band (akin to a variable in dataset)
    band = f.GetRasterBand(1)
    # get the pixel width, height, etc.
    transform = f.GetGeoTransform()
    # convert lat,lon to row,col
    column = (Long - transform[0]) / transform[1]
    row = (Lat - transform[3]) / transform[5]
    # get pixel values surrounding data point
    Surrounding_data=(band.ReadAsArray(np.floor(column-2), np.floor(row-2), 5, 5))
    # convert row,col back to north,east
    Long_c = transform[0] + np.floor(column) * res
    Lat_c = transform[3] - np.floor(row) * res
    # set up matrices for interpolation
    count=-1
    pos=np.zeros((25,2))
    Surrounding_data_v=np.zeros((25,1))
    for k in range(-2,3):
        for j in range(-2,3):
            count=count+1
            pos[count]=(Long_c+j*res,Lat_c-k*res)
            Surrounding_data_v[count]=Surrounding_data[k+2,j+2]
    interp_g=griddata(pos,Surrounding_data_v,(Long,Lat))
    return interp_g

def normal_correction(Lat_A,Long_A,H_A,Lat_B,Long_B,H_B):
    # ellipsoidal gravity at 45 deg. Lat
    Gamma_0=9.8061992115
    # Normal Gravity at the point
    normal_g_A=mean_normal_grav(Lat_A,H_A)
#    print normal_g_A
    normal_g_B=mean_normal_grav(Lat_B,H_B)
#    print normal_g_B
    dn=H_B-H_A
    g=mean_surface_grav(Lat_A,Long_A,H_A,Lat_B,Long_B,H_B)
   # print g
    NC=(dn*(g-Gamma_0)/Gamma_0)+H_A*(normal_g_A-Gamma_0)/Gamma_0-H_B*(normal_g_B-Gamma_0)/Gamma_0
    return NC,g


def normal_orthometric_correction(lat1, lon1, H1, lat2, lon2, H2):
    """
    Computes the normal-orthometric correction based on Heck (2003).
    See Standard for New Zealand Vertical Datum 2016, Section 3.3

    :param lat1: Latitude at Stn1
    :param lon1: Longitude at Stn1
    :param H1: Physical Height at Stn1
    :param lat2: Latitude at Stn2
    :param lon2: longitude at Stn2
    :param H2: Physical Height at Stn2
    :return: normal-orthometric correction
    """

    f_ng = cons.grs80_ngf
    m_rad = cons.grs80.meanradius

    mid_height = (H1 + H2) / 2
    mid_lat = m.radians((lat1 + lat2) / 2)

    vinc_inv = gg.vincinv(lat1, lon1, lat2, lon2)
    dist = vinc_inv[0]
    az = vinc_inv[1]

    noc = - f_ng / m_rad * mid_height * m.sin(2.0 * mid_lat) * m.cos(m.radians(az)) * dist

    return noc



#!/usr/bin/env python3

"""
Geoscience Australia - Python Geodesy Package
Coordinate Module
"""

from geodepy.constants import Projection, utm, grs80
from geodepy.angles import (DECAngle, HPAngle, GONAngle, DMSAngle, DDMAngle,
                            dec2hpa, dec2gona, dec2dms, dec2ddm,
                            angular_typecheck)
from geodepy.convert import xyz2llh, llh2xyz, grid2geo, geo2grid


class CoordCart(object):
    """
    Cartesian Coordinate Class
    Used for working with coordinates representing points in a 3 dimensional
    earth-centred earth-fixed system. N Value (optional) is the distance (m)
    at the coordinate that a reference surface (typically a geoid) is above
    or below an ellipsoid
    """
    def __init__(self, xaxis, yaxis, zaxis, nval=None):
        """
        :param xaxis: Distance (m) along X Axis (positive when passing through
        intersection of equator and prime meridian)
        :type xaxis: float
        :param yaxis: Distance (m) along Y Axis (positive when passing through
        intersection of equator and 90 degrees east longitude)
        :type yaxis: float
        :param zaxis: Distance (m) along Z Axis (positive when passing through
        north pole)
        :type zaxis: float
        :param nval: Distance (m) between a reference surface (typically a
        geoid) and an ellipsoid (positive when surface is above ellipsoid)
        :type nval: float
        """
        self.xaxis = float(xaxis)
        self.yaxis = float(yaxis)
        self.zaxis = float(zaxis)
        if not nval:
            self.nval = None
        else:
            self.nval = float(nval)

    def __repr__(self):
        return (f'CoordCart: X: {self.xaxis} Y: {self.yaxis} Z: {self.zaxis} '
                f'NVal: {self.nval}')

    def __eq__(self, other):
        if isinstance(other, CoordCart):
            return vars(self) == vars(other)
        else:
            raise ValueError(f"Can't compare {self} to {other}. If other object"
                             f" is coord, convert to CoordCart first.")

    def __round__(self, n=None):
        if self.nval is None:
            return CoordCart(round(self.xaxis, n), round(self.yaxis, n),
                             round(self.zaxis, n))
        else:
            return CoordCart(round(self.xaxis, n), round(self.yaxis, n),
                             round(self.zaxis, n), round(self.nval, n))

    def geo(self, ellipsoid=grs80, notation=DECAngle):
        """
        Convert coordinates to Geographic
        Note: If no N Value, no Orthometric Height set.
        :param ellipsoid: geodepy.constants.Ellipsoid Object (default: grs80)
        :param notation: Latitude and Longitude Angle Notation format
        :type notation: geodepy.angle class or float
        :return: Geographic Coordinate
        :rtype: CoordGeo
        """
        lat, lon, ell_ht = xyz2llh(self.xaxis, self.yaxis,
                                   self.zaxis, ellipsoid)
        if notation is DECAngle:
            lat = DECAngle(lat)
            lon = DECAngle(lon)
        elif notation is HPAngle:
            lat = DECAngle(lat).hpa()
            lon = DECAngle(lon).hpa()
        elif notation is GONAngle:
            lat = DECAngle(lat).gona()
            lon = DECAngle(lon).gona()
        elif notation is DMSAngle:
            lat = DECAngle(lat).dms()
            lon = DECAngle(lon).dms()
        elif notation is DDMAngle:
            lat = DECAngle(lat).ddm()
            lon = DECAngle(lon).ddm()
        elif notation is float:
            pass  # geodepy.convert.grid2geo returns float dec degrees
        else:
            raise ValueError(f'CoordCart.geo() notation requires class float or'
                             f' class from geodepy.angles module. '
                             f'Supplied: {notation}')
        if self.nval is None:
            return CoordGeo(lat, lon, ell_ht)
        else:
            return CoordGeo(lat, lon, ell_ht,
                            ell_ht - self.nval)

    # TODO: Add functionality to utilise different TM projections

    def tm(self, ellipsoid=grs80, projection=utm):
        """
        Convert coordinates to Transverse Mercator
        :param ellipsoid: geodepy.constants.Ellipsoid Object (default: grs80)
        :param projection: geodepy.constants.Projection Object (default: utm)
        :return: Transverse Mercator Projection Coordinate
        :rtype: CoordTM
        """
        return self.geo(ellipsoid).tm(ellipsoid, projection)


class CoordGeo(object):
    """
    Geographic Coordinate Class
    Used for working with coordinates representing points in an ellipsoidal
    system with ellipsoid heights (m) relative to the surface of the ellipsoid.
    Orthometric heights (m) are relative to a different reference surface
    (typically a geoid).
    """
    def __init__(self, lat, lon, ell_ht=None, orth_ht=None):
        """
        :param lat: Geographic Latitude (angle between +90 and -90 degrees,
        positive values are north of equator)
        :type lat: float (decimal degrees) or any Angle object in geodepy.angles
        :param lon: Geographic Longitude (angle between +180 and -180 degrees,
        positive values are east of central meridian)
        :type lon: float (decimal degrees) or any Angle object in geodepy.angles
        :param ell_ht: Ellipsoid Height (m, positive is outside ellipsoid)
        :type ell_ht: float
        :param orth_ht: Orthometric Height (m)
        :type orth_ht: float
        """
        # Check latitude and longitude are supported types, both same type
        type_list = [float, DECAngle, HPAngle, GONAngle, DMSAngle, DDMAngle]
        if not all(x in type_list for x in [type(lat), type(lon)]):
            raise TypeError(f'CoordGeo Latitude and Longitude must be type: '
                            f'float (decimal degrees) or geodepy.angles class '
                            f'Lat: {type(lat)}, Lon: {type(lon)}')
        if type(lat) != type(lon):
            raise TypeError(f'CoordGeo Latitude and Longitude must have the '
                            f'same notation type: float (decimal degrees) or '
                            f'geodepy.angles class. Lat: {type(lat)}, '
                            f'Lon: {type(lon)}')
        self.lat = lat
        self.lon = lon
        if ell_ht is None:
            self.ell_ht = None
        else:
            self.ell_ht = float(ell_ht)
        if orth_ht is None:
            self.orth_ht = None
        else:
            self.orth_ht = float(orth_ht)

    def __repr__(self):
        return (f'CoordGeo: Lat: {self.lat} Lon: {self.lon} '
                f'Ell_Ht: {self.ell_ht} Orth_Ht: {self.orth_ht}')

    def __eq__(self, other):
        if isinstance(other, CoordGeo):
            return vars(self) == vars(other)
        else:
            raise ValueError(f"Can't compare {self} to {other}. If other object"
                             f" is coord, convert to CoordGeo first.")

    def __round__(self, n=None):
        if self.ell_ht is None and self.orth_ht is None:
            return CoordGeo(round(self.lat, n), round(self.lon, n))
        elif self.ell_ht is None:
            return CoordGeo(round(self.lat, n), round(self.lon, n),
                            None, round(self.orth_ht, n))
        elif self.orth_ht is None:
            return CoordGeo(round(self.lat, n), round(self.lon, n),
                            round(self.ell_ht, n), None)
        else:
            return CoordGeo(round(self.lat, n), round(self.lon, n),
                            round(self.ell_ht, n), round(self.orth_ht, n))

    def notation(self, notation):
        if type(self.lat) == float:  # Decimal Degrees (float)
            # Use functions to convert from Decimal Degrees (float)
            if notation == float:
                pass
            elif notation == DECAngle:
                new_lat = DECAngle(self.lat)
                new_lon = DECAngle(self.lon)
            elif notation == HPAngle:
                new_lat = dec2hpa(self.lat)
                new_lon = dec2hpa(self.lon)
            elif notation == GONAngle:
                new_lat = dec2gona(self.lat)
                new_lon = dec2gona(self.lon)
            elif notation == DMSAngle:
                new_lat = dec2dms(self.lat)
                new_lon = dec2dms(self.lon)
            elif notation == DDMAngle:
                new_lat = dec2ddm(self.lat)
                new_lon = dec2ddm(self.lon)
            else:
                raise ValueError(
                    f'CoordGeo.notation() notation requires class float or '
                    f'class from geodepy.angles module. '
                    f'Supplied: {notation}')
        elif type(self.lat) in [DECAngle, HPAngle, GONAngle,
                                DMSAngle, DDMAngle]:
            # Use methods to convert from geodepy.angles classes
            if notation == float:
                new_lat = self.lat.dec()
                new_lon = self.lon.dec()
            elif notation == DECAngle:
                new_lat = self.lat.deca()
                new_lon = self.lon.deca()
            elif notation == HPAngle:
                new_lat = self.lat.hpa()
                new_lon = self.lon.hpa()
            elif notation == GONAngle:
                new_lat = self.lat.gona()
                new_lon = self.lon.gona()
            elif notation == DMSAngle:
                new_lat = self.lat.dms()
                new_lon = self.lon.dms()
            elif notation == DDMAngle:
                new_lat = self.lat.ddm()
                new_lon = self.lon.ddm()
            else:
                raise ValueError(
                    f'CoordGeo.notation() notation requires class float or '
                    f'class from geodepy.angles module. '
                    f'Supplied: {notation}')
        return CoordGeo(new_lat, new_lon, self.ell_ht, self.orth_ht)

    def cart(self, ellipsoid=grs80):
        """
        Convert coordinates to Cartesian
        Note: If no ellipsoid height set, uses 0m. No N Value output
        :param ellipsoid: geodepy.constants.Ellipsoid Object (default: grs80)
        :return: Cartesian Coordinate
        :rtype: CoordCart
        """
        if self.ell_ht:
            x, y, z = llh2xyz(self.lat, self.lon, self.ell_ht, ellipsoid)
            if self.orth_ht:  # Only N Value if both Ellipsoid and Ortho Heights
                return CoordCart(x, y, z, self.ell_ht - self.orth_ht)
            else:  # No Ortho Height -> No N Value
                return CoordCart(x, y, z)
        else:  # No Ellipsoid Height - set to 0m for conversion
            x, y, z = llh2xyz(self.lat, self.lon, 0, ellipsoid)
            return CoordCart(x, y, z)  # No Ellipsoid Height -> No N Value

    # TODO: Add functionality to utilise different TM projections

    def tm(self, ellipsoid=grs80, projection=utm):
        """
        Convert coordinates to Universal Transverse Mercator Projection
        Note: Heights are not used in calculation
        :param ellipsoid: geodepy.constants.Ellipsoid Object (default: grs80)
        :param projection: geodepy.constants.Projection Object (default: utm)
        :return: Transverse Mercator Projection Coordinate
        :rtype: CoordTM
        """
        hemi, zone, east, north, psf, gc = geo2grid(self.lat, self.lon,
                                                    0, ellipsoid)
        if hemi == 'North':
            hemi_north = True
        else:  # hemi == 'South'
            hemi_north = False

        return CoordTM(zone, east, north,
                       self.ell_ht, self.orth_ht,
                       hemi_north, projection)


class CoordTM(object):
    """
    Transverse Mercator Coordinate Class
    Used for working with coordinates representing points in a Transverse
    Mercator projected planar system. Coordinates are related to an ellipsoid
    via a projection (default is Universal Transverse Mercator, see
    geodepy.constants.Projection for more details). Ellipsoid heights (m)
    are relative to the surface of the ellipsoid. Orthometric heights (m) are
    relative to a different reference surface (typically a geoid).
    """
    def __init__(self, zone, east, north, ell_ht=None, orth_ht=None,
                 hemi_north=False, projection=utm):
        """
        :param zone: Transverse Mercator Zone Number - 1 to 60
        :type zone: int
        :param east: Easting (m, within 3330km of Central Meridian)
        :type east: float
        :param north: Northing (m, 0 to 10,000,000m)
        :type north: float
        :param ell_ht: Ellipsoid Height (m, positive is outside ellipsoid)
        :type ell_ht: float
        :param orth_ht: Orthometric Height (m)
        :type orth_ht: float
        :param hemi_north: True if coordinate in Northern Hemisphere, False if
        coordinate in Southern Hemisphere (default)
        :type hemi_north: bool
        :param projection: Information defining relationship between projected
        coordinate and the ellipsoid
        :type projection: geodepy.constants.Projection object
        (default: geodepy.constants.utm)
        """
        self.zone = int(zone)
        self.east = float(east)
        self.north = float(north)
        if ell_ht is None:
            self.ell_ht = None
        else:
            self.ell_ht = float(ell_ht)
        if orth_ht is None:
            self.orth_ht = None
        else:
            self.orth_ht = float(orth_ht)
        if isinstance(hemi_north, bool):
            self.hemi_north = hemi_north
        else:
            raise TypeError(f'CoordTM: invalid hemi_north, must be bool (True '
                            f'for Northern Hemisphere, False for Southern). '
                            f'hemi_north entered: {hemi_north}')
        if isinstance(projection, Projection):
            self.projection = projection
        else:
            raise TypeError(f'CoordTM: invalid projection, must be geodepy.'
                            f'constants.Projection object (default is geodepy.'
                            f'constants.utm. projection entered: {projection}')

    def __repr__(self):
        if self.hemi_north:
            return (f'CoordTM: Zone: {self.zone} East: {self.east} '
                    f'North: {self.north} Ell_Ht: {self.ell_ht} '
                    f'Orth_Ht: {self.orth_ht} Hemisphere: North')
        else:  # not self.hemi_north
            return (f'CoordTM: Zone: {self.zone} East: {self.east} '
                    f'North: {self.north} Ell_Ht: {self.ell_ht} '
                    f'Orth_Ht: {self.orth_ht} Hemisphere: South')

    def __eq__(self, other):
        if isinstance(other, CoordTM):
            return vars(self) == vars(other)
        else:
            raise ValueError(f"Can't compare {self} to {other}. If other object"
                             f" is coord, convert to CoordTM first.")

    def __round__(self, n=None):
        if self.ell_ht is None and self.orth_ht is None:
            return CoordTM(self.zone, round(self.east, n), round(self.north, n),
                           None, None,
                           self.hemi_north, self.projection)
        elif self.ell_ht is None:
            return CoordTM(self.zone, round(self.east, n), round(self.north, n),
                           None, round(self.orth_ht, n),
                           self.hemi_north, self.projection)
        elif self.orth_ht is None:
            return CoordTM(self.zone, round(self.east, n), round(self.north, n),
                           round(self.ell_ht, n), None,
                           self.hemi_north, self.projection)
        else:
            return CoordTM(self.zone, round(self.east, n), round(self.north, n),
                           round(self.ell_ht, n), round(self.orth_ht, n),
                           self.hemi_north, self.projection)

    def geo(self, ellipsoid=grs80, notation=DECAngle):
        """

        :param ellipsoid: geodepy.constants.Ellipsoid Object (default: grs80)
        :param notation: Latitude and Longitude Angle Notation format
        :type notation: geodepy.angle class or float
        :return:
        """
        if self.hemi_north:
            hemi_str = 'north'
        else:
            hemi_str = 'south'
        lat, lon, psf, grid_conv = grid2geo(self.zone, self.east, self.north,
                                            hemi_str, ellipsoid)
        if notation is DECAngle:
            lat = DECAngle(lat)
            lon = DECAngle(lon)
        elif notation is HPAngle:
            lat = DECAngle(lat).hpa()
            lon = DECAngle(lon).hpa()
        elif notation is GONAngle:
            lat = DECAngle(lat).gona()
            lon = DECAngle(lon).gona()
        elif notation is DMSAngle:
            lat = DECAngle(lat).dms()
            lon = DECAngle(lon).dms()
        elif notation is DDMAngle:
            lat = DECAngle(lat).ddm()
            lon = DECAngle(lon).ddm()
        elif notation is float:
            pass  # geodepy.convert.grid2geo returns float dec degrees
        else:
            raise ValueError(f'CoordTM.geo() notation requires class float or '
                             f'class from geodepy.angles module. '
                             f'Supplied: {notation}')
        return CoordGeo(lat, lon, self.ell_ht, self.orth_ht)

    # TODO: Add functionality to utilise different TM projections

    def cart(self, ellipsoid=grs80):
        """
        Convert coordinates to Cartesian
        Note: If no ellipsoid height set, uses 0m. No N Value output
        :param ellipsoid: geodepy.constants.Ellipsoid Object (default: grs80)
        :return: Cartesian Coordinate
        :rtype: CoordCart
        """
        return self.geo(ellipsoid).cart(ellipsoid)



"""
GeodePy - Python Geodesy Package
ntv2reader Module

Adapted from Jaimie Dodd's ntv2reader.py for AUSGeoid ntv2 files (2018-11-12).
Expanded (2021-03-21) for the more general case to include horizontal transformation grids
"""

import struct
from datetime import datetime as dt
import numpy as np


class NTv2Grid(object):
    def __init__(self, num_orec, num_srec, num_file, gs_type, version, system_f, system_t, major_f, minor_f, major_t,
                 minor_t, file_path):
        self.num_orec = num_orec        # Number of header identifiers
        self.num_srec = num_srec        # Number of sub-header idents
        self.num_file = num_file        # Number of subgrids in file
        self.gs_type = gs_type          # grid shift type
        self.version = version          # grid file version
        self.system_f = system_f        # reference system
        self.system_t = system_t        # reference system
        self.major_f = major_f          # semi major of from system
        self.minor_f = minor_f          # semi minor of from system
        self.major_t = major_t          # semi major of to system
        self.minor_t = minor_t          # semi minor of to system
        self.subgrids = {}
        self.file_path = file_path


class SubGrid(object):
    def __init__(self, sub_name, parent, created, updated, s_lat, n_lat, e_long, w_long, lat_inc, long_inc, gs_count):
        self.sub_name = sub_name        # subgrid name
        self.parent = parent            # parent name
        self.created = created          # date created
        self.updated = updated          # date modified
        self.s_lat = s_lat              # south latitude extents
        self.n_lat = n_lat              # north latitude extents
        self.e_long = e_long            # east longitude extents
        self.w_long = w_long            # west longitude extents
        self.lat_inc = lat_inc          # latitude increment
        self.long_inc = long_inc        # longitude increment
        self.gs_count = gs_count        # total nodes in subgrid

    def ntv2_bilinear(self, lat, lon, num_cols, row, col, f, start_byte):
        """
        Function to perform bicubic interpolation of the four ntv2 fields at a point of interest.

        :param lat: latitude of the point of interest (arc-seconds)
        :param lon: longitude of the point of interest (negative arc-seconds)
        :param num_cols: number of columns in grid
        :param row: row number of point to the bottom right of the point of interest
        :param col: column number of point to the bottom right of the point of interest
        :param f: variable for open file NTv2 being read as binary
        :param start_byte: start index of subgrid

        :return: four field tuple of interpolation results at point of interest.
        """

        #   |     |
        # --o-----o--
        #   |4    |3
        #   | *P  |
        # --o-----o--
        #   |2    |1
        #
        # o - Node position.

        # Determine position in the file of the four surrounding nodes
        pos1 = row * num_cols + col
        pos2 = pos1 + 1
        pos3 = pos1 + num_cols
        pos4 = pos3 + 1

        # Navigate to start of subgrid
        f.seek(start_byte, 1)
        # Navigate to start of posA node
        f.seek(16 * pos1, 1)

        # Read in values for nodes 1 and 2
        node_1 = read_node(f)
        node_2 = read_node(f)

        # Navigate to beginning of node 3
        f.seek(16*(pos3 - pos2 - 1), 1)

        # Read in values for nodes 3 and 4
        node_3 = read_node(f)
        node_4 = read_node(f)

        # Determine latitude and longitude of node 1
        lat1 = self.s_lat + row * self.lat_inc
        long1 = self.e_long + col * self.long_inc

        # Determine interpolation scale factors
        x = (lon - long1) / self.long_inc
        y = (lat - lat1) / self.lat_inc

        field_1 = round(bilinear_interpolation(node_1[0], node_2[0], node_3[0], node_4[0], x, y), 6)
        field_2 = round(bilinear_interpolation(node_1[1], node_2[1], node_3[1], node_4[1], x, y), 6)
        field_3 = round(bilinear_interpolation(node_1[2], node_2[2], node_3[2], node_4[2], x, y), 6)
        field_4 = round(bilinear_interpolation(node_1[3], node_2[3], node_3[3], node_4[3], x, y), 6)

        return field_1, field_2, field_3, field_4

    def ntv2_bicubic(self, lat, lon, num_cols, row, col, f, start_byte):
        """
        Function to perform bicubic interpolation of the four ntv2 fields at a point of interest.

        :param lat: latitude of the point of interest (arc-seconds)
        :param lon: longitude of the point of interest (negative arc-seconds)
        :param num_cols: number of columns in grid
        :param row: row number of point to the bottom right of the point of interest
        :param col: column number of point to the bottom right of the point of interest
        :param f: variable for open file NTv2 being read as binary
        :param start_byte: start index of subgrid

        :return: four field tuple of interpolation results at point of interest.

        """

        #   |     |     |     |
        # --o-----o-----o-----o--
        #   |11   |12   |13   |14
        #   |     |     |     |
        # --o-----o-----o-----o--
        #   |10   |3    |4    |15
        #   |     | *P  |     |
        # --o-----o-----o-----o--
        #   |9    |2    |1    |16
        #   |     |     |     |
        # --o-----o-----o-----o--
        #   |8    |7    |6    |5
        #
        # o - Node position.

        # Determine position in the file of the sixteen surrounding nodes
        pos1 = row * num_cols + col
        pos2 = pos1 + 1
        pos3 = pos2 + num_cols
        pos4 = pos3 - 1
        pos5 = pos4 - 2 * num_cols - 1
        pos6 = pos5 + 1
        pos7 = pos6 + 1
        pos8 = pos7 + 1
        pos9 = pos8 + num_cols
        pos10 = pos9 + num_cols
        pos11 = pos10 + num_cols
        pos12 = pos11 - 1
        pos13 = pos12 - 1
        pos14 = pos13 - 1
        pos15 = pos14 - num_cols
        pos16 = pos15 - num_cols

        # Navigate to start of subgrid
        f.seek(start_byte, 1)
        # Navigate to start of pos1 node
        f.seek(16 * pos5, 1)

        # Read in values for nodes 5-8
        node_5 = read_node(f)
        node_6 = read_node(f)
        node_7 = read_node(f)
        node_8 = read_node(f)

        # Navigate to start of pos16 node
        f.seek(16 * (pos16 - pos8 - 1), 1)

        # Read in values for nodes 16, 1, 2, and 9
        node_16 = read_node(f)
        node_1 = read_node(f)
        node_2 = read_node(f)
        node_9 = read_node(f)

        # Navigate to start of pos15 node
        f.seek(16 * (pos15 - pos9 - 1), 1)

        # Read in values for nodes 15, 3, 4 and 10
        node_15 = read_node(f)
        node_4 = read_node(f)
        node_3 = read_node(f)
        node_10 = read_node(f)

        # Navigate to start of pos14 node
        f.seek(16 * (pos14 - pos10 - 1), 1)

        # Read in values for nodes 11, 12, 13 and 14
        node_14 = read_node(f)
        node_13 = read_node(f)
        node_12 = read_node(f)
        node_11 = read_node(f)

        # Determine latitude and longitude of node 1
        lat1 = self.s_lat + row * self.lat_inc
        long1 = self.e_long + col * self.long_inc

        # Determine interpolation scale factors
        x = round((lon - long1) / self.long_inc, 6)
        y = round((lat - lat1) / self.lat_inc, 6)

        # Call bicubic interpolation function to interpolate ntv2 fields.
        field_1 = round(bicubic_interpolation(node_1[0], node_2[0], node_3[0], node_4[0],
                                              node_5[0], node_6[0], node_7[0], node_8[0],
                                              node_9[0], node_10[0], node_11[0], node_12[0],
                                              node_13[0], node_14[0], node_15[0], node_16[0],
                                              x, y), 6)
        field_2 = round(bicubic_interpolation(node_1[1], node_2[1], node_3[1], node_4[1],
                                              node_5[1], node_6[1], node_7[1], node_8[1],
                                              node_9[1], node_10[1], node_11[1], node_12[1],
                                              node_13[1], node_14[1], node_15[1], node_16[1],
                                              x, y), 6)
        field_3 = round(bicubic_interpolation(node_1[2], node_2[2], node_3[2], node_4[2],
                                              node_5[2], node_6[2], node_7[2], node_8[2],
                                              node_9[2], node_10[2], node_11[2], node_12[2],
                                              node_13[2], node_14[2], node_15[2], node_16[2],
                                              x, y), 6)
        field_4 = round(bicubic_interpolation(node_1[3], node_2[3], node_3[3], node_4[3],
                                              node_5[3], node_6[3], node_7[3], node_8[3],
                                              node_9[3], node_10[3], node_11[3], node_12[3],
                                              node_13[3], node_14[3], node_15[3], node_16[3],
                                              x, y), 6)

        return field_1, field_2, field_3, field_4


def read_node(f):
    """
    Function to read in values of nodes

    :param f: variable for open file NTv2 being read as binary
    :return: tuple containing the four ntv2 fields at the grid node.
    """

    # field_1: shift lat / geoid separation (N)
    byte = f.read(4)
    field_1 = struct.unpack('f', byte)[0]
    # field 2: shift lon / deflection in prime meridian value (Xi)
    byte = f.read(4)
    field_2 = struct.unpack('f', byte)[0]
    # field 3: reliability lat / deflection in prime vertical value (Eta)
    byte = f.read(4)
    field_3 = struct.unpack('f', byte)[0]
    # field 4: reliability lon / NA
    byte = f.read(4)
    field_4 = struct.unpack('f', byte)[0]

    return field_1, field_2, field_3, field_4


def bilinear_interpolation(n1, n2, n3, n4, x, y):
    """
    Bilinear interpolation of value for point of interest (P).

    :param n1: value at node 1
    :param n2: value at node 2
    :param n3: value at node 3
    :param n4: value at node 4
    :param x: interpolation scale factor for x axis
    :param y: interpolation scale factor for y axis

    :return: value at node P
    """

    a0 = n1
    a1 = n2 - n1
    a2 = n3 - n1
    a3 = n1 + n4 - n2 - n3
    p = a0 + (a1 * x) + (a2 * y) + (a3 * x * y)
    return p


def bicubic_interpolation(n1, n2, n3, n4, n5, n6, n7, n8, n9, n10, n11, n12, n13, n14, n15, n16, x, y):
    """
    Bicubic interpolation of value for point of interest (P).

    :param n1: value at node 1
    :param n2: value at node 2
    :param n3: value at node 3
    :param n4: value at node 4
    :param n5: value at node 5
    :param n6: value at node 6
    :param n7: value at node 7
    :param n8: value at node 8
    :param n9: value at node 9
    :param n10: value at node 10
    :param n11: value at node 11
    :param n12: value at node 12
    :param n13: value at node 13
    :param n14: value at node 14
    :param n15: value at node 16
    :param n16: value at node 17
    :param x: interpolation scale factor for x axis
    :param y: interpolation scale factor for y axis

    :return: value at node P
    """

    # Define the inverse of the coefficient matrix
    cinv = np.array([[1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
                     [0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0],
                     [-3, 0, 0, 3, 0, 0, 0, 0, -2, 0, 0, -1, 0, 0, 0, 0],
                     [2, 0, 0, -2, 0, 0, 0, 0, 1, 0, 0, 1, 0, 0, 0, 0],
                     [0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
                     [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0],
                     [0, 0, 0, 0, -3, 0, 0, 3, 0, 0, 0, 0, -2, 0, 0, -1],
                     [0, 0, 0, 0, 2, 0, 0, -2, 0, 0, 0, 0, 1, 0, 0, 1],
                     [-3, 3, 0, 0, -2, -1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
                     [0, 0, 0, 0, 0, 0, 0, 0, -3, 3, 0, 0, -2, -1, 0, 0],
                     [9, -9, 9, -9, 6, 3, -3, -6, 6, -6, -3, 3, 4, 2, 1, 2],
                     [-6, 6, -6, 6, -4, -2, 2, 4, -3, 3, 3, -3, -2, -1, -1, -2],
                     [2, -2, 0, 0, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
                     [0, 0, 0, 0, 0, 0, 0, 0, 2, -2, 0, 0, 1, 1, 0, 0],
                     [-6, 6, -6, 6, -3, -3, 3, 3, -4, 4, 2, -2, -2, -2, -1, -1],
                     [4, -4, 4, -4, 2, 2, -2, -2, 2, -2, -2, 2, 1, 1, 1, 1]])

    # Define x parameters
    # Function values
    x1 = n1
    x2 = n2
    x3 = n3
    x4 = n4
    # X Derivatives
    x5 = (n2 - n16) / 2
    x6 = (n9 - n1) / 2
    x7 = (n10 - n4) / 2
    x8 = (n3 - n15) / 2
    # Y Derivatives
    x9 = (n4 - n6) / 2
    x10 = (n3 - n7) / 2
    x11 = (n12 - n2) / 2
    x12 = (n13 - n1) / 2
    # Cross Derivatives (XY)
    x13 = (n3 - n7 - n15 + n5) / 4
    x14 = (n10 - n8 - n4 + n6) / 4
    x15 = (n11 - n9 - n13 + n1) / 4
    x16 = (n12 - n2 - n14 + n16) / 4

    # Create array from x parameters
    xarr = np.array([x1, x2, x3, x4, x5, x6, x7, x8, x9, x10, x11, x12, x13, x14, x15, x16])

    # Multiply the inverse of the coefficient matrix by the array of x values to give array of alpha values
    alpha = np.matmul(cinv, xarr)

    # Calculate value at the point of interest
    n_p = 0
    for i in range(0, 4):
        for j in range(0, 4):
            n_p = n_p + alpha[i * 4 + j] * x ** i * y ** j

    return n_p


def read_ntv2_file(ntv2_gsb_file):
    """
    Function to read an ntv2 gsb file and create grid & subgrid objects

    :param ntv2_gsb_file: full path to ntv2 file
    :return: ntv2 grid object
    """

    with open(ntv2_gsb_file, 'rb') as f:
        # NUM_OREC
        f.seek(8, 1)
        byte = f.read(4)
        num_orec = int.from_bytes(byte, byteorder='little')

        f.seek(4, 1)

        # NUM_SREC
        f.seek(8, 1)
        byte = f.read(4)
        num_srec = int.from_bytes(byte, byteorder='little')

        f.seek(4, 1)
        # NUM_FILE
        f.seek(8, 1)
        byte = f.read(4)
        num_file = int.from_bytes(byte, byteorder='little')

        f.seek(4, 1)
        # GS_TYPE
        f.seek(8, 1)
        byte = f.read(8)
        gs_type = byte.decode('utf8').strip('\x00').strip()

        # VERSION
        f.seek(8, 1)
        byte = f.read(8)
        version = byte.decode('utf8').strip('\x00').strip()

        # SYSTEM_F
        f.seek(8, 1)
        byte = f.read(8)
        system_f = byte.decode('utf8').strip('\x00').strip()

        # SYSTEM_T
        f.seek(8, 1)
        byte = f.read(8)
        system_t = byte.decode('utf8').strip('\x00').strip()

        # MAJOR_F
        f.seek(8, 1)
        byte = f.read(8)
        major_f = struct.unpack('d', byte)[0]

        # MINOR_F
        f.seek(8, 1)
        byte = f.read(8)
        minor_f = struct.unpack('d', byte)[0]

        # MAJOR_T
        f.seek(8, 1)
        byte = f.read(8)
        major_t = struct.unpack('d', byte)[0]

        # MINOR_T
        f.seek(8, 1)
        byte = f.read(8)
        minor_t = struct.unpack('d', byte)[0]

        grid = NTv2Grid(
            num_orec=num_orec,
            num_srec=num_srec,
            num_file=num_file,
            gs_type=gs_type,
            version=version,
            system_f=system_f,
            system_t=system_t,
            major_f=major_f,
            minor_f=minor_f,
            major_t=major_t,
            minor_t=minor_t,
            file_path=ntv2_gsb_file
        )

        # read subgrids
        for i in range(0, grid.num_file):
            # SUB_NAME
            f.seek(8, 1)
            byte = f.read(8)
            sub_name = byte.decode('utf').strip('\x00').strip()

            # PARENT
            f.seek(8, 1)
            byte = f.read(8)
            parent = byte.decode('utf').strip('\x00').strip()

            # CREATED
            f.seek(8, 1)
            byte = f.read(8)
            created = dt.strptime(byte.decode('utf').strip('\x00').strip(), '%d%m%Y').strftime('%d/%m/%Y')

            # UPDATED
            f.seek(8, 1)
            byte = f.read(8)
            updated = dt.strptime(byte.decode('utf').strip('\x00').strip(), '%d%m%Y').strftime('%d/%m/%Y')

            # S_LAT
            f.seek(8, 1)
            byte = f.read(8)
            s_lat = round(struct.unpack('d', byte)[0], 3)

            # N_LAT
            f.seek(8, 1)
            byte = f.read(8)
            n_lat = round(struct.unpack('d', byte)[0], 3)

            # E_LONG
            f.seek(8, 1)
            byte = f.read(8)
            e_long = round(struct.unpack('d', byte)[0], 3)

            # W_LONG
            f.seek(8, 1)
            byte = f.read(8)
            w_long = round(struct.unpack('d', byte)[0], 3)

            # LAT_INC
            f.seek(8, 1)
            byte = f.read(8)
            lat_inc = round(struct.unpack('d', byte)[0], 6)

            # LONG_INC
            f.seek(8, 1)
            byte = f.read(8)
            long_inc = round(struct.unpack('d', byte)[0], 6)

            # GS_COUNT
            f.seek(8, 1)
            byte = f.read(4)
            gs_count = int.from_bytes(byte, byteorder='little')

            f.seek(4, 1)
            # skip ahead to next subgrid
            f.seek(16 * gs_count, 1)

            grid.subgrids[sub_name] = SubGrid(
                sub_name=sub_name,
                parent=parent,
                created=created,
                updated=updated,
                s_lat=s_lat,
                n_lat=n_lat,
                e_long=e_long,
                w_long=w_long,
                lat_inc=lat_inc,
                long_inc=long_inc,
                gs_count=gs_count
            )

        return grid


def interpolate_ntv2(grid_object, lat, lon, method='bicubic'):
    """
    Function to interpolate Ntv2Grid objects
    :param grid_object: Ntv2Grid object
    :param lat: latitude (decimal degrees)
    :param lon: longitude (decimal degrees)
    :param method: interpolation strategy, bicubic or bilinear

    :return: tuple of four ntv2 fields
    """

    interpolation_methods = {
        'bicubic',
        'bilinear'
    }
    if method not in interpolation_methods:
        raise ValueError(f'interpolation method "{method}" not supported')

    # convert decimal degrees to arc-seconds
    lat *= 3600
    lon *= -3600

    # determine subgrid for point of interest
    in_subgrids = set()
    for sg in grid_object.subgrids.values():
        if sg.s_lat <= lat < sg.n_lat and sg.e_long <= lon < sg.w_long:
            in_subgrids.add(sg.sub_name)

    # return null fields if no subgrid found, else choose subgrid with finest resolution
    if len(in_subgrids) == 0:
        return None, None, None, None
    else:
        inc = None
        in_grid = None
        for sg in in_subgrids:
            if not inc:
                inc = grid_object.subgrids[sg].lat_inc
                in_grid = grid_object.subgrids[sg]
            else:
                if grid_object.subgrids[sg].lat_inc < inc:
                    inc = grid_object.subgrids[sg].lat_inc
                    in_grid = grid_object.subgrids[sg]

    # Determine number of columns in grid, and row and column of node to bottom right of
    # point of interest, then call relevant interpolation method function

    # determine number of columns
    num_cols = 1 + int((in_grid.w_long - in_grid.e_long) / in_grid.long_inc)

    # determine row and col numbers of node below right of point
    row = int((lat - in_grid.s_lat) / in_grid.lat_inc)
    col = int((lon - in_grid.e_long) / in_grid.long_inc)

    # locate data in gsb_file
    skip_bytes = 176    # grid header length

    with open(grid_object.file_path, 'rb') as f:
        for sg in grid_object.subgrids.values():

            skip_bytes += 176   # subgrid header length
            if sg.sub_name == in_grid.sub_name:
                if method == 'bilinear':
                    results = in_grid.ntv2_bilinear(lat, lon, num_cols, row, col, f, skip_bytes)
                elif method == 'bicubic':
                    results = in_grid.ntv2_bicubic(lat, lon, num_cols, row, col, f, skip_bytes)
            else:
                skip_bytes += sg.gs_count * 16

    return results
